# Copyright (c) 2013-2014 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Convert labeled graph to dot using
pydot and custom filtering
"""
from __future__ import division
import logging
import re
from collections import Iterable
from textwrap import fill
from cStringIO import StringIO
import numpy as np
import networkx as nx
from networkx.utils import make_str
import pydot
# inline:
#
# import webcolors


logger = logging.getLogger(__name__)


def _states2dot_str(graph, to_pydot_graph, wrap=10,
                    tikz=False, rankdir='TB'):
    """Copy nodes to given Pydot graph, with attributes for dot export."""
    # TODO generate LaTeX legend table for edge labels

    states = graph.states

    # get labeling def
    if hasattr(graph, '_state_label_def'):
        label_def = graph._state_label_def

    if hasattr(graph, '_state_dot_label_format'):
        label_format = graph._state_dot_label_format
    else:
        label_format = {'type?label': '', 'separator': r'\\n'}

    for u, d in graph.nodes_iter(data=True):
        # initial state ?
        is_initial = u in states.initial
        is_accepting = _is_accepting(graph, u)

        # state annotation
        node_dot_label = _form_node_label(
            u, d, label_def,
            label_format, wrap, tikz=tikz
        )

        # node_dot_label = fill(str(state), width=wrap)

        rim_color = d.get('color', 'black')

        if tikz:
            _state2tikz(graph, to_pydot_graph, u,
                        is_initial, is_accepting, rankdir,
                        rim_color, d, node_dot_label)
        else:
            _state2dot(graph, to_pydot_graph, u,
                       is_initial, is_accepting,
                       rim_color, d, node_dot_label)


def _state2dot(graph, to_pydot_graph, state,
               is_initial, is_accepting,
               rim_color, d, node_dot_label):
    if is_initial:
        _add_incoming_edge(to_pydot_graph, state)

    normal_shape = graph.dot_node_shape['normal']
    accept_shape = graph.dot_node_shape.get('accepting', '')

    shape = accept_shape if is_accepting else normal_shape
    corners = 'rounded' if shape is 'rectangle' else ''

    rim_color = '"' + _format_color(rim_color, 'dot') + '"'

    fc = d.get('fillcolor', 'none')
    filled = '' if fc is 'none' else 'filled'
    if fc is 'gradient':
        # top/bottom colors not supported for dot

        lc = d.get('left_color', d['top_color'])
        rc = d.get('right_color', d['bottom_color'])

        if isinstance(lc, basestring):
            fillcolor = lc
        elif isinstance(lc, dict):
            fillcolor = lc.keys()[0]
        else:
            raise TypeError('left_color must be str or dict.')

        if isinstance(rc, basestring):
            fillcolor += ':' + rc
        elif isinstance(rc, dict):
            fillcolor += ':' + rc.keys()[0]
        else:
            raise TypeError('right_color must be str or dict.')
    else:
        fillcolor = _format_color(fc, 'dot')

    if corners and filled:
        node_style = '"' + corners + ', ' + filled + '"'
    elif corners:
        node_style = '"' + corners + '"'
    else:
        node_style = '"' + filled + '"'

    to_pydot_graph.add_node(
        state,
        label=node_dot_label,
        shape=shape,
        style=node_style,
        color=rim_color,
        fillcolor='"' + fillcolor + '"')


def _state2tikz(graph, to_pydot_graph, state,
                is_initial, is_accepting, rankdir,
                rim_color, d, node_dot_label):
    style = 'state'

    if rankdir is 'LR':
        init_dir = 'initial left'
    elif rankdir is 'RL':
        init_dir = 'initial right'
    elif rankdir is 'TB':
        init_dir = 'initial above'
    elif rankdir is 'BT':
        init_dir = 'initial below'
    else:
        raise ValueError('Unknown rankdir')

    if is_initial:
        style += ', initial by arrow, ' + init_dir + ', initial text='
    if is_accepting:
        style += ', accepting'

    if graph.dot_node_shape['normal'] is 'rectangle':
        style += ', shape = rectangle, rounded corners'

    # darken the rim
    if 'black' in rim_color:
        c = _format_color(rim_color, 'tikz')
    else:
        c = _format_color(rim_color, 'tikz') + '!black!30'

    style += ', draw = ' + c

    fill = d.get('fillcolor')

    if fill is 'gradient':
        s = {'top_color', 'bottom_color',
             'left_color', 'right_color'}
        for x in s:
            if x in d:
                style += ', ' + x + ' = ' + _format_color(d[x], 'tikz')
    elif fill is not None:
        # not gradient
        style += ', fill = ' + _format_color(fill, 'tikz')
    else:
        logger.debug('fillcolor is None')

    to_pydot_graph.add_node(
        state,
        texlbl=node_dot_label,
        style=style)


def _format_color(color, prog='tikz'):
    """Encode color in syntax for given program.

    @type color:
      - C{str} for single color or
      - C{dict} for weighted color mix

    @type prog: 'tikz' or 'dot'
    """
    if isinstance(color, basestring):
        return color

    if not isinstance(color, dict):
        raise Exception('color must be str or dict')

    if prog is 'tikz':
        s = '!'.join([k + '!' + str(v) for k, v in color.iteritems()])
    elif prog is 'dot':
        t = sum(color.itervalues())

        try:
            import webcolors

            # mix them
            result = np.array((0.0, 0.0, 0.0))
            for c, w in color.iteritems():
                result += w/t * np.array(webcolors.name_to_rgb(c))
            s = webcolors.rgb_to_hex(result)
        except:
            logger.warn('failed to import webcolors')
            s = ':'.join([k + ';' + str(v/t) for k, v in color.iteritems()])
    else:
        raise ValueError('Unknown program: ' + str(prog) + '. '
                         "Available options are: 'dot' or 'tikz'.")
    return s


def _place_initial_states(trs_graph, pd_graph, tikz):
    init_subg = pydot.Subgraph('initial')
    init_subg.set_rank('source')

    for node in trs_graph.states.initial:
        pd_node = pydot.Node(make_str(node))
        init_subg.add_node(pd_node)

        phantom_node = 'phantominit' + str(node)
        pd_node = pydot.Node(make_str(phantom_node))
        init_subg.add_node(pd_node)

    pd_graph.add_subgraph(init_subg)


def _add_incoming_edge(g, state):
    phantom_node = 'phantominit' + str(state)
    g.add_node(phantom_node, label='""', shape='none', width='0')
    g.add_edge(phantom_node, state)


def _form_node_label(state, state_data, label_def,
                     label_format, width=10, tikz=False):
    # node itself
    state_str = str(state)
    state_str = state_str.replace("'", "")

    # rm parentheses to reduce size of states in fig
    if tikz:
        state_str = state_str.replace('(', '')
        state_str = state_str.replace(')', '')

    # make indices subscripts
    if tikz:
        pattern = '([a-zA-Z]\d+)'
        make_subscript = lambda x: x.group(0)[0] + '_' + x.group(0)[1:]
        state_str = re.sub(pattern, make_subscript, state_str)

    # SVG requires breaking the math environment into
    # one math env per line. Just make 1st line math env
    # if latex:
    #    state_str = '$' + state_str + '$'
    #    state_str = fill(state_str, width=width)
    node_dot_label = state_str

    # newline between state name and label, only if state is labeled
    if len(state_data) != 0:
        node_dot_label += r'\\n'

    # add node annotations from action, AP sets etc
    # other key,values in state attr_dict ignored
    pieces = list()
    for (label_type, label_value) in state_data.iteritems():
        if label_type not in label_def:
            continue

        # label formatting
        type_name = label_format[label_type]
        sep_type_value = label_format['type?label']

        # avoid turning strings to lists,
        # or non-iterables to lists
        if isinstance(label_value, str):
            label_str = fill(label_value, width=width)
        elif isinstance(label_value, Iterable):  # and not str
            s = ', '.join([str(x) for x in label_value])
            label_str = r'\\{' + fill(s, width=width) + r'\\}'
        else:
            label_str = fill(str(label_value), width=width)

        pieces.append(type_name + sep_type_value + label_str)

    sep_label_sets = label_format['separator']
    node_dot_label += sep_label_sets.join(pieces)

    if tikz:
        # replace LF by latex newline
        node_dot_label = node_dot_label.replace(r'\\n', r'\\\\ ')

        # dot2tex math mode doesn't handle newlines properly
        node_dot_label = (
            r'$\\begin{matrix} ' + node_dot_label +
            r'\\end{matrix}$'
        )

    return node_dot_label


def _is_accepting(graph, state):
    """accepting state ?"""
    # no accepting states defined ?
    if not hasattr(graph.states, 'accepting'):
        return False

    return state in graph.states.accepting


def _transitions2dot_str(trans, to_pydot_graph, tikz=False):
    """Convert transitions to dot str.

    @rtype: str
    """
    if not hasattr(trans.graph, '_transition_label_def'):
        return
    if not hasattr(trans.graph, '_transition_dot_label_format'):
        return
    if not hasattr(trans.graph, '_transition_dot_mask'):
        return

    # get labeling def
    label_def = trans.graph._transition_label_def
    label_format = trans.graph._transition_dot_label_format
    label_mask = trans.graph._transition_dot_mask

    for (u, v, key, edge_data) in trans.graph.edges_iter(
        data=True, keys=True
    ):
        edge_dot_label = _form_edge_label(
            edge_data, label_def,
            label_format, label_mask, tikz
        )

        edge_color = edge_data.get('color', 'black')

        to_pydot_graph.add_edge(u, v, key=key,
                                label=edge_dot_label,
                                color=edge_color)


def _form_edge_label(edge_data, label_def,
                     label_format, label_mask, tikz):
    label = ''  # dot label for edge
    sep_label_sets = label_format['separator']

    for label_type, label_value in edge_data.iteritems():
        if label_type not in label_def:
            continue

        # masking defined ?
        # custom filter hiding based on value
        if label_type in label_mask:
            # not show ?
            if not label_mask[label_type](label_value):
                continue

        # label formatting
        if label_type in label_format:
            type_name = label_format[label_type]
            sep_type_value = label_format['type?label']
        else:
            type_name = ':'
            sep_type_value = r',\\n'

        # format iterable containers using
        # mathematical set notation: {...}
        if isinstance(label_value, basestring):
            # str is Iterable: avoid turning it to list
            label_str = label_value
        elif isinstance(label_value, Iterable):
            s = ', '.join([str(x) for x in label_value])
            label_str = r'\\{' + fill(s) + r'\\}'
        else:
            label_str = str(label_value)

        if tikz:
            type_name = r'\mathrm' + '{' + type_name + '}'

        label += (type_name + sep_type_value +
                  label_str + sep_label_sets)

    if tikz:
        label = r'\\begin{matrix}' + label + r'\\end{matrix}'

    label = '"' + label + '"'

    return label


def _graph2pydot(graph, wrap=10, tikz=False,
                 rankdir='TB'):
    """Convert (possibly labeled) state graph to dot str.

    @type graph: L{LabeledDiGraph}

    @rtype: str
    """
    dummy_nx_graph = nx.MultiDiGraph()

    _states2dot_str(graph, dummy_nx_graph, wrap=wrap, tikz=tikz,
                    rankdir=rankdir)
    _transitions2dot_str(graph.transitions, dummy_nx_graph, tikz=tikz)

    pydot_graph = nx.drawing.nx_pydot.to_pydot(dummy_nx_graph)
    _place_initial_states(graph, pydot_graph, tikz)

    pydot_graph.set_overlap('false')
    # pydot_graph.set_size('"0.25,1"')
    # pydot_graph.set_ratio('"compress"')
    pydot_graph.set_nodesep(0.5)
    pydot_graph.set_ranksep(0.1)

    return pydot_graph


def graph2dot_str(graph, wrap=10, tikz=False):
    """Convert graph to dot string.

    Requires pydot.

    @type graph: L{LabeledDiGraph}

    @param wrap: textwrap width

    @rtype: str
    """
    pydot_graph = _graph2pydot(graph, wrap=wrap, tikz=tikz)

    return pydot_graph.to_string()


def save_dot(graph, path, fileformat, rankdir, prog, wrap, tikz=False):
    """Save state graph to dot file.

    @type graph: L{LabeledDiGraph}

    @return: True upon success
    @rtype: bool
    """
    pydot_graph = _graph2pydot(graph, wrap=wrap, tikz=tikz,
                               rankdir=rankdir)
    if pydot_graph is None:
        # graph2dot must have printed warning already
        return False
    pydot_graph.set_rankdir(rankdir)
    pydot_graph.set_splines('true')

    # turn off graphviz warnings caused by tikz labels
    if tikz:
        prog = [prog, '-q 1']
    pydot_graph.write(path, format=fileformat, prog=prog)
    return True


def plot_pydot(graph, prog='dot', rankdir='LR', wrap=10, ax=None):
    """Plot a networkx or pydot graph using dot.

    No files written or deleted from the disk.

    Note that all networkx graph classes are inherited
    from networkx.Graph

    See Also
    ========
    dot & pydot documentation

    @param graph: to plot
    @type graph: networkx.Graph | pydot.Graph

    @param prog: GraphViz programto use
    @type prog: 'dot' | 'neato' | 'circo' | 'twopi'
        | 'fdp' | 'sfdp' | etc

    @param rankdir: direction to layout nodes
    @type rankdir: 'LR' | 'TB'

    @param ax: axes
    """
    try:
        pydot_graph = _graph2pydot(graph, wrap=wrap)
    except:
        if isinstance(graph, nx.Graph):
            pydot_graph = nx.drawing.nx_pydot.to_pydot(graph)
        else:
            raise TypeError(
                'graph not networkx or pydot class.' +
                'Got instead: ' + str(type(graph)))
    pydot_graph.set_rankdir(rankdir)
    pydot_graph.set_splines('true')
    pydot_graph.set_bgcolor('gray')

    png_str = pydot_graph.create_png(prog=prog)

    # installed ?
    try:
        from IPython.display import display, Image
        logger.debug('IPython installed.')

        # called by IPython ?
        try:
            cfg = get_ipython().config
            logger.debug('Script called by IPython.')

            # Caution!!! : not ordinary dict,
            # but IPython.config.loader.Config

            # qtconsole ?
            if cfg['IPKernelApp']:
                logger.debug('Within IPython QtConsole.')
                display(Image(data=png_str))
                return True
        except:
            print('IPython installed, but not called from it.')
    except ImportError:
        logger.warn('IPython not found.\nSo loaded dot images not inline.')

    # not called from IPython QtConsole, try Matplotlib...

    # installed ?
    try:
        import matplotlib.pyplot as plt
        import matplotlib.image as mpimg
    except:
        logger.debug('Matplotlib not installed.')
        logger.warn('Neither IPython QtConsole nor Matplotlib available.')
        return None

    logger.debug('Matplotlib installed.')

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    sio = StringIO()
    sio.write(png_str)
    sio.seek(0)
    img = mpimg.imread(sio)
    ax.imshow(img, aspect='equal')
    plt.show(block=False)

    return ax



"""Tests for the tulip.gridworld."""
import numpy as np
import tulip.gridworld as gw
from tulip.synth import is_realizable


REFERENCE_GWFILE = """
# A very small example, realizable by itself.
6 10
*  G*
  ***  ***
         *
I  *  *  *
  ****** *
*
"""

UNREACHABLE_GOAL_GWFILE = """
4 4
**G
  **

I* *
"""

TRIVIAL_GWFILE = """
2 2
*
"""


# Module-level fixture setup
def setUp():
    np.random.seed(0)  # Make pseudorandom number sequence repeatable


class GridWorld_test(object):
    def setUp(self):
        self.prefix = "testworld"
        self.X = gw.GridWorld(REFERENCE_GWFILE, prefix=self.prefix)
        self.Y_testpaths = gw.GridWorld(UNREACHABLE_GOAL_GWFILE,
                                        prefix=self.prefix)

    def tearDown(self):
        self.X = None
        self.Y_testpaths = None

    def test_reachability(self):
        # Reachability is assumed to be bidirectional
        assert not self.Y_testpaths.is_reachable((3,0), (0,2))
        assert not self.Y_testpaths.is_reachable((0,2), (3,0))
        assert self.Y_testpaths.is_reachable((1,1), (2,3))
        assert self.Y_testpaths.is_reachable((2,3), (1,1))

    def test_size(self):
        assert self.X.size() == (6, 10)

    def test_copy(self):
        Z = self.X.copy()
        assert Z is not self.X
        assert Z.W is not self.X.W
        assert Z == self.X

    def test_getitem(self):
        assert self.X.__getitem__(
            (0,0), nonbool=False) == self.prefix+"_"+str(0)+"_"+str(0)
        assert self.X.__getitem__(
            (-1,0), nonbool=False) == self.prefix+"_"+str(5)+"_"+str(0)
        assert self.X.__getitem__(
            (-1,-2), nonbool=False) == self.prefix+"_"+str(5)+"_"+str(8)

    def test_state(self):
        assert self.X.state((2,3), nonbool=False) == {
            'testworld_3_9': 0, 'testworld_1_8': 0, 'testworld_1_9': 0,
            'testworld_1_4': 0, 'testworld_1_5': 0, 'testworld_1_6': 0,
            'testworld_1_7': 0, 'testworld_1_0': 0, 'testworld_1_1': 0,
            'testworld_1_2': 0, 'testworld_1_3': 0, 'testworld_0_5': 0,
            'testworld_0_4': 0, 'testworld_0_7': 0, 'testworld_0_6': 0,
            'testworld_0_1': 0, 'testworld_0_0': 0, 'testworld_0_3': 0,
            'testworld_0_2': 0, 'testworld_5_7': 0, 'testworld_0_9': 0,
            'testworld_0_8': 0, 'testworld_3_2': 0, 'testworld_3_3': 0,
            'testworld_2_9': 0, 'testworld_2_8': 0, 'testworld_3_6': 0,
            'testworld_3_7': 0, 'testworld_3_4': 0, 'testworld_3_5': 0,
            'testworld_2_3': 1, 'testworld_2_2': 0, 'testworld_2_1': 0,
            'testworld_2_0': 0, 'testworld_2_7': 0, 'testworld_2_6': 0,
            'testworld_2_5': 0, 'testworld_2_4': 0, 'testworld_4_1': 0,
            'testworld_4_0': 0, 'testworld_4_3': 0, 'testworld_4_2': 0,
            'testworld_4_5': 0, 'testworld_4_4': 0, 'testworld_4_7': 0,
            'testworld_4_6': 0, 'testworld_4_9': 0, 'testworld_4_8': 0,
            'testworld_5_8': 0, 'testworld_5_2': 0, 'testworld_5_9': 0,
            'testworld_3_0': 0, 'testworld_3_1': 0, 'testworld_5_3': 0,
            'testworld_5_5': 0, 'testworld_5_0': 0, 'testworld_5_4': 0,
            'testworld_5_1': 0, 'testworld_5_6': 0, 'testworld_3_8': 0}
        assert self.X.state((-1,0), nonbool=False) == {
            'testworld_3_9': 0, 'testworld_1_8': 0, 'testworld_1_9': 0,
            'testworld_1_4': 0, 'testworld_1_5': 0, 'testworld_1_6': 0,
            'testworld_1_7': 0, 'testworld_1_0': 0, 'testworld_1_1': 0,
            'testworld_1_2': 0, 'testworld_1_3': 0, 'testworld_0_5': 0,
            'testworld_0_4': 0, 'testworld_0_7': 0, 'testworld_0_6': 0,
            'testworld_0_1': 0, 'testworld_0_0': 0, 'testworld_0_3': 0,
            'testworld_0_2': 0, 'testworld_5_7': 0, 'testworld_0_9': 0,
            'testworld_0_8': 0, 'testworld_3_2': 0, 'testworld_3_3': 0,
            'testworld_2_9': 0, 'testworld_2_8': 0, 'testworld_3_6': 0,
            'testworld_3_7': 0, 'testworld_3_4': 0, 'testworld_3_5': 0,
            'testworld_2_3': 0, 'testworld_2_2': 0, 'testworld_2_1': 0,
            'testworld_2_0': 0, 'testworld_2_7': 0, 'testworld_2_6': 0,
            'testworld_2_5': 0, 'testworld_2_4': 0, 'testworld_4_1': 0,
            'testworld_4_0': 0, 'testworld_4_3': 0, 'testworld_4_2': 0,
            'testworld_4_5': 0, 'testworld_4_4': 0, 'testworld_4_7': 0,
            'testworld_4_6': 0, 'testworld_4_9': 0, 'testworld_4_8': 0,
            'testworld_5_8': 0, 'testworld_5_2': 0, 'testworld_5_9': 0,
            'testworld_3_0': 0, 'testworld_3_1': 0, 'testworld_5_3': 0,
            'testworld_5_5': 0, 'testworld_5_0': 1, 'testworld_5_4': 0,
            'testworld_5_1': 0, 'testworld_5_6': 0, 'testworld_3_8': 0}
        assert self.X.state((-1,-1), nonbool=False) == {
            'testworld_3_9': 0, 'testworld_1_8': 0, 'testworld_1_9': 0,
            'testworld_1_4': 0, 'testworld_1_5': 0, 'testworld_1_6': 0,
            'testworld_1_7': 0, 'testworld_1_0': 0, 'testworld_1_1': 0,
            'testworld_1_2': 0, 'testworld_1_3': 0, 'testworld_0_5': 0,
            'testworld_0_4': 0, 'testworld_0_7': 0, 'testworld_0_6': 0,
            'testworld_0_1': 0, 'testworld_0_0': 0, 'testworld_0_3': 0,
            'testworld_0_2': 0, 'testworld_5_7': 0, 'testworld_0_9': 0,
            'testworld_0_8': 0, 'testworld_3_2': 0, 'testworld_3_3': 0,
            'testworld_2_9': 0, 'testworld_2_8': 0, 'testworld_3_6': 0,
            'testworld_3_7': 0, 'testworld_3_4': 0, 'testworld_3_5': 0,
            'testworld_2_3': 0, 'testworld_2_2': 0, 'testworld_2_1': 0,
            'testworld_2_0': 0, 'testworld_2_7': 0, 'testworld_2_6': 0,
            'testworld_2_5': 0, 'testworld_2_4': 0, 'testworld_4_1': 0,
            'testworld_4_0': 0, 'testworld_4_3': 0, 'testworld_4_2': 0,
            'testworld_4_5': 0, 'testworld_4_4': 0, 'testworld_4_7': 0,
            'testworld_4_6': 0, 'testworld_4_9': 0, 'testworld_4_8': 0,
            'testworld_5_8': 0, 'testworld_5_2': 0, 'testworld_5_9': 1,
            'testworld_3_0': 0, 'testworld_3_1': 0, 'testworld_5_3': 0,
            'testworld_5_5': 0, 'testworld_5_0': 0, 'testworld_5_4': 0,
            'testworld_5_1': 0, 'testworld_5_6': 0, 'testworld_3_8': 0}

    def test_equality(self):
        assert self.X == gw.GridWorld(REFERENCE_GWFILE)
        Y = gw.GridWorld()
        assert self.X != Y
        Y = gw.GridWorld(TRIVIAL_GWFILE)
        assert self.X != Y

    def test_dumploadloop(self):
        assert self.X == gw.GridWorld(self.X.dumps())

    def test_spec_realizable_bool(self):
        spec = self.X.spec(nonbool=False)
        spec.moore = False
        spec.plus_one = False
        spec.qinit = r'\A \E'
        assert is_realizable('omega', spec)

    def test_spec_realizable(self):
        spec = self.X.spec()
        spec.moore = False
        spec.plus_one = False
        spec.qinit = r'\A \E'
        assert is_realizable('omega', spec)

    def check_is_empty(self, coord, expected):
        assert self.X.is_empty(coord) == expected

    def test_is_empty(self):
        for coord, expected in [((0, 0), False), ((0, 1), True),
                                ((-1, 0), False), ((0, -1), True)]:
            yield self.check_is_empty, coord, expected

    def check_is_empty_extend(self, coord, expected):
        assert self.X.is_empty(coord, extend=True) == expected

    def test_is_empty_extend(self):
        for coord, expected in [((0, 0), False), ((0, 1), True),
                                ((-1, 0), False), ((0, -1), False)]:
            yield self.check_is_empty_extend, coord, expected

    def test_dump_subworld(self):
        # No offset
        X_local = self.X.dump_subworld((2,4), prefix="X")
        assert X_local.size() == (2, 4)
        assert X_local.__getitem__((0,0), nonbool=False) == "X_0_0"
        assert not X_local.is_empty((0,0))
        assert X_local.is_empty((0,1))

        # Offset
        X_local = self.X.dump_subworld((2,4), offset=(1,0), prefix="Xoff")
        assert X_local.size() == (2, 4)
        assert X_local.__getitem__((0,0), nonbool=False) == "Xoff_0_0"
        assert X_local.is_empty((0,0))
        assert X_local.is_empty((0,1))
        assert not X_local.is_empty((0,3))

    def test_dump_subworld_extend(self):
        # No offset
        Xsize = self.X.size()
        X_local = self.X.dump_subworld((Xsize[0]+1, Xsize[1]), prefix="X",
                                      extend=True)
        X_local.goal_list = self.X.goal_list[:]
        X_local.init_list = self.X.init_list[:]
        assert X_local.size() == (7, 10)
        assert X_local.__getitem__((0,0), nonbool=False) == "X_0_0"
        assert not X_local.is_empty((0,0))
        assert X_local.is_empty((0,1))
        # Equal except for the last row, which should be all occupied in X_local
        X_local_s = X_local.dumps().splitlines()
        assert np.all(X_local_s[1:-1] == self.X.dumps().splitlines()[1:])
        assert not X_local.is_empty((6,1))
        assert X_local_s[-1] == "*"*10

        # Offset
        X_local = self.X.dump_subworld((3,4), offset=(-1,0), prefix="Xoff",
                                      extend=True)
        assert X_local.size() == (3, 4)
        assert X_local.__getitem__((0,0), nonbool=False) == "Xoff_0_0"
        assert not X_local.is_empty((0,0))
        assert not X_local.is_empty((0,1))
        assert not X_local.is_empty((0,3))
        assert X_local.is_empty((1,1))


class RandomWorld_test(object):
    def setUp(self):
        self.wall_densities = [.2, .4, .6]
        self.sizes = [(4,5), (4,5), (10,20)]
        self.rworlds = [
            gw.random_world(
                self.sizes[r], wall_density=self.wall_densities[r], prefix="Y")
            for r in range(len(self.sizes))]
        self.rworlds_ensuredfeasible = [
            gw.random_world(
                self.sizes[r], self.wall_densities[r], num_init=2,
                num_goals=2, ensure_feasible=True)
            for r in range(len(self.sizes))]

    def tearDown(self):
        self.rworlds = []

    def test_feasibility(self):
        for r in range(len(self.rworlds_ensuredfeasible)):
            print "test \"ensured feasible\" world index", r
            print self.rworlds_ensuredfeasible[r]
            assert self.rworlds_ensuredfeasible[r].is_reachable(
                self.rworlds_ensuredfeasible[r].init_list[0],
                self.rworlds_ensuredfeasible[r].init_list[1])
            assert self.rworlds_ensuredfeasible[r].is_reachable(
                self.rworlds_ensuredfeasible[r].init_list[1],
                self.rworlds_ensuredfeasible[r].goal_list[0])
            assert self.rworlds_ensuredfeasible[r].is_reachable(
                self.rworlds_ensuredfeasible[r].goal_list[0],
                self.rworlds_ensuredfeasible[r].goal_list[1])
            assert self.rworlds_ensuredfeasible[r].is_reachable(
                self.rworlds_ensuredfeasible[r].goal_list[1],
                self.rworlds_ensuredfeasible[r].init_list[0])

    def test_size(self):
        for r in range(len(self.rworlds)):
            print "test world index", r
            print self.rworlds[r]
            assert self.sizes[r] == self.rworlds[r].size()

    def test_density(self):
        for r in range(len(self.rworlds)):
            print "test world index", r
            print self.rworlds[r]
            (num_rows, num_cols) = self.rworlds[r].size()
            num_occupied = 0
            for i in range(num_rows):
                for j in range(num_cols):
                    if not self.rworlds[r].is_empty((i,j)):
                        num_occupied += 1
            assert (
                float(num_occupied) / (num_rows*num_cols) ==
                self.wall_densities[r])

RandomWorld_test.slow = True


def extract_coord_check(label, expected_coord):
    assert gw.extract_coord(label) == expected_coord


def extract_coord_test():
    for (label, expected_coord) in [
            ("test_3_0", ("test", 3, 0)),
            ("obstacle_5_4_11", ("obstacle_5", 4, 11)),
            ("test3_0", None)]:
        yield extract_coord_check, label, expected_coord


def eq_gridworld_check(G, H, eq):
    if eq:
        G == H
    else:
        not (G == H)


def eq_gridworld_test():
    empty = gw.GridWorld()
    trivial_nonempty = gw.GridWorld(TRIVIAL_GWFILE)
    trivial_diff = gw.GridWorld(TRIVIAL_GWFILE)
    if trivial_diff.is_empty((0, 0)):
        trivial_diff.mark_occupied((0, 0))
    else:
        trivial_diff.mark_empty((0, 0))
    trivial_nonempty_2goals = gw.GridWorld(TRIVIAL_GWFILE)
    trivial_nonempty_2goals.goal_list = [(0, 0), (1, 1)]
    trivial_nonempty_2init = gw.GridWorld(TRIVIAL_GWFILE)
    trivial_nonempty_2init.init_list = [(0, 0), (1, 1)]
    for (G, H, is_equal) in [
            (gw.GridWorld(), gw.GridWorld(), True),
            (empty, trivial_nonempty, False),
            (trivial_nonempty_2goals, trivial_nonempty, False),
            (trivial_nonempty_2init, trivial_nonempty, False),
            (trivial_nonempty, trivial_diff, False),
            (gw.unoccupied((3, 5)), gw.unoccupied((1, 1)), False)]:
        yield eq_gridworld_check, G, H, is_equal


def narrow_passage_test():
    G = gw.narrow_passage((5, 10), num_init=1, num_goals=1)
    assert G.is_reachable(G.init_list[0], G.goal_list[0])


def scale_gridworld_test():
    G = gw.unoccupied((1, 2))
    assert G.size() == (1, 2)
    assert G.scale().size() == G.size()
    assert G.scale(xf=1, yf=1).size() == G.size()
    assert G.scale(xf=2).size() == (1, 4)
    assert G.scale(yf=2).size() == (2, 2)
    assert G.scale(xf=3, yf=4).size() == (4, 6)


def add_trolls_test():
    G = gw.unoccupied((3, 5))
    G.init_list = [(0, 0)]
    G.goal_list = [(0, 4)]
    spc = gw.add_trolls(G, [((2, 2), 1)], get_moves_lists=False)
    spc.moore = False
    spc.plus_one = False
    spc.qinit = r'\A \E'
    assert is_realizable('omega', spc)



"""
This example illustrates how to export a Mealy Machine to
Matlab/Simulink/Stateflow (see the very last line).

For detailed comments on the example itself,
please see examples/robot_planning/discrete.py
"""
from tulip import transys, spec, synth

# import file that contains to_stateflow
import sys
sys.path.append('../')
import tomatlab

# Create a finite transition system
sys = transys.FTS()

# Define the states of the system
sys.states.add_from(['X0', 'X1', 'X2', 'X3', 'X4', 'X5'])
sys.states.initial.add('X0')    # start in state X0

# Define the allowable transitions
sys.transitions.add_comb({'X0'}, {'X1', 'X3'})
sys.transitions.add_comb({'X1'}, {'X0', 'X4', 'X2'})
sys.transitions.add_comb({'X2'}, {'X1', 'X5'})
sys.transitions.add_comb({'X3'}, {'X0', 'X4'})
sys.transitions.add_comb({'X4'}, {'X3', 'X1', 'X5'})
sys.transitions.add_comb({'X5'}, {'X4', 'X2'})

# Add atomic propositions to the states
sys.atomic_propositions.add_from({'home', 'lot'})
sys.states.add('X0', ap={'home'})
sys.states.add('X5', ap={'lot'})

# Environment variables and specification
env_vars = {'park'}
env_init = set()
env_prog = '!park'
env_safe = set()

# System specification
sys_vars = {'X0reach'}
sys_init = {'X0reach'}
sys_prog = {'home'}
sys_safe = {'(X (X0reach) <-> lot) || (X0reach && !park)'}
sys_prog |= {'X0reach'}

# Create the specification
specs = spec.GRSpec(env_vars, sys_vars, env_init, sys_init,
                    env_safe, sys_safe, env_prog, sys_prog)

# Controller synthesis
ctrl = synth.synthesize('gr1c', specs, sys=sys)

# Generate a MATLAB script that generates a Mealy Machine
tomatlab.export('robot_discrete.mat', ctrl)



import polytope
import numpy
from poly2str import matrix2str, polytope2str


def pppexport(ppp, filename="pppdata"):
    """Export a proposition preserving partition to a C file

    Restrictions:
        - Each region is assumed to be a convex polytope.
        - If not, then only the first polytope inside the region is
          exported. Hence, the MPC may be infeasible

    @param ppp: tulip.abstract.prop2partition.PropPreservingPartition
    @param filename: a string of the export filename
    @rtype: None
    """
    numregions = len(ppp.regions)
    # generate the .c file
    f = open("sources/"+filename+".c", "w")

    f.write('#include "'+filename+'.h"\n\n')

    f.write("idxint nRegion = "+str(numregions)+";\n")
    f.write("Polytope* regions["+str(numregions)+"];\n")

    for index in range(0, numregions):
        region = ppp.regions[index]
        p = region.list_poly[0]
        s = polytope2str(p, 'p'+str(index))
        f.write("/****Polytope "+str(index) + " ****/\n")
        f.write(s+"\n")

    f.write("void init_region()\n{\n")
    for index in range(0, numregions):
        f.write("   regions[" + str(index) + "]=create_poly(p" + str(index) +
                "k,p" + str(index) + "l,p" + str(index) + "A,p" + str(index) +
                "b,p" + str(index) + "center);\n")

    f.write("}\n\n")

    f.write("""\
void free_region()\n{
    int i;
    for (i=0;i<nRegion;i++)
    {
        free(regions[i]);
    }
}""")

    f.close()



"""
Tests for tulip.hybrid module
"""
import numpy as np

from nose.tools import raises
from tulip import hybrid
import polytope as pc

def switched_system_test():
    subsystems = []

    # subsystem 0
    A = np.eye(2)
    B = np.eye(2)

    Uset = pc.box2poly([[0.0, 1.0], [0.0, 1.0]])
    domain0 = pc.box2poly([[0.0, 2.0], [0.0, 2.0]])

    subsystems += [hybrid.LtiSysDyn(A, B, Uset=Uset, domain=domain0)]

    # subsystem 1
    domain1 = pc.box2poly([[2.0, 4.0], [0.0, 2.0]])

    subsystems += [hybrid.LtiSysDyn(A, B, Uset=Uset, domain=domain1)]

    # PWA system
    domain = domain0.union(domain1)
    pwa = hybrid.PwaSysDyn(subsystems, domain)

    # Switched system (mode dynamics the same, just testing code)
    dom = (2, 2)
    dyn = {
        ('a', 'c'):pwa,
        ('a', 'd'):pwa,
        ('b', 'c'):pwa,
    }
    env_labels = ['a', 'b']
    sys_labels = ['c', 'd']

    hyb = hybrid.SwitchedSysDyn(
        disc_domain_size=dom,
        dynamics=dyn,
        cts_ss=domain,
        env_labels=env_labels,
        disc_sys_labels=sys_labels
    )

    print(hyb)

    assert(hyb.disc_domain_size == dom)
    assert(hyb.dynamics == dyn)
    assert(hyb.env_labels == env_labels)
    assert(hyb.disc_sys_labels == sys_labels)
    assert(hyb.cts_ss == domain)


class time_semantics_test:
    """Test out time semantics for hybrid systems module."""
    def setUp(self):
        self.A1 = np.eye(2)
        self.A2 = np.array([[0, 1], [0, 0]])
        self.B1 = np.array([[0] ,[1]])
        self.B2 = np.array([[1], [0]])
        self.poly1 = pc.Polytope.from_box([[0, 1], [0, 1]])
        self.poly2 = pc.Polytope.from_box([[1, 2], [0, 1]])
        self.total_box = pc.Region(list_poly=[self.poly1, self.poly2])
        self.Uset = pc.Polytope.from_box([[0, 1]])

    def tearDown(self):
        self.A1 = None
        self.A2 = None
        self.B1 = None
        self.B2 = None
        self.poly1 = None
        self.poly2 = None
        self.total_box = None
        self.Uset = None

    def test_correct_lti_construction(self):
        LTI1 = hybrid.LtiSysDyn(A=self.A1, B=self.B1,
                                Uset=self.Uset, domain=self.poly1,
                                time_semantics='discrete', timestep=None)
        LTI2 = hybrid.LtiSysDyn(A=self.A2, B=self.B2,
                                Uset=self.Uset, domain=self.poly2,
                                time_semantics='sampled', timestep=.1)
        LTI3 = hybrid.LtiSysDyn(A=self.A2, B=self.B2,
                                Uset=self.Uset, domain=self.poly1)
        LTI4 = hybrid.LtiSysDyn(A=self.A1, B=self.B2,
                                Uset=self.Uset, domain=self.poly2)
        assert(LTI1.time_semantics == 'discrete')
        assert(LTI2.time_semantics == 'sampled')
        assert(LTI1.timestep is None)
        assert(LTI2.timestep == .1)
        assert(LTI3.time_semantics is None)
        assert(LTI3.timestep is None)
        assert(LTI4.time_semantics is None)
        assert(LTI4.timestep is None)

    def test_correct_pwa_construction(self):
        # Putting pwa together successfully, without time overwrite
        LTI1 = hybrid.LtiSysDyn(A=self.A1, B=self.B1,
                                Uset=self.Uset, domain=self.poly1,
                                time_semantics='sampled', timestep=.1)
        LTI2 = hybrid.LtiSysDyn(A=self.A2, B=self.B2,
                                Uset=self.Uset, domain=self.poly2,
                                time_semantics='sampled', timestep=.1)
        LTI3 = hybrid.LtiSysDyn(A=self.A2, B=self.B2,
                                Uset=self.Uset, domain=self.poly1)
        LTI4 = hybrid.LtiSysDyn(A=self.A1, B=self.B2,
                                Uset=self.Uset, domain=self.poly2)
        PWA1 = hybrid.PwaSysDyn(list_subsys=[LTI3, LTI4], domain=self.total_box,
                                overwrite_time=False)

        # Putting pwa together successfully, with time overwrite
        PWA2 = hybrid.PwaSysDyn(list_subsys=[LTI1, LTI2], domain=self.total_box,
                                time_semantics='sampled', timestep=.1,
                                overwrite_time=True)

    @raises(ValueError)
    def test_pwa_difftseman_among_subsys(self):
        """Different time semantics among LtiSysDyn subsystems of PwaSysDyn"""
        LTI1 = hybrid.LtiSysDyn(A=self.A1, B=self.B1,
                                Uset=self.Uset, domain=self.poly1,
                                time_semantics='sampled', timestep=.1)
        LTI2 = hybrid.LtiSysDyn(A=self.A2, B=self.B2,
                                Uset=self.Uset, domain=self.poly2,
                                time_semantics='discrete')
        PWA = hybrid.PwaSysDyn(list_subsys=[LTI1, LTI2], domain=self.total_box,
                               time_semantics='sampled', timestep=.1,
                               overwrite_time=False)

    @raises(ValueError)
    def test_pwa_difftstep_among_subsys(self):
        """Different timesteps among LtiSysDyn subsystems of PwaSysDyn"""
        LTI1 = hybrid.LtiSysDyn(A=self.A1, B=self.B1,
                                Uset=self.Uset, domain=self.poly1,
                                time_semantics='sampled', timestep=.1)
        LTI2 = hybrid.LtiSysDyn(A=self.A2, B=self.B2,
                                Uset=self.Uset, domain=self.poly2,
                                time_semantics='sampled', timestep=.2)
        PWA = hybrid.PwaSysDyn(list_subsys=[LTI1, LTI2], domain=self.total_box,
                               time_semantics='sampled', timestep=.1,
                               overwrite_time=False)

    @raises(ValueError)
    def test_pwa_difftseman_from_subsys(self):
        """LtiSysDyn subsystems time semantics do not match that of PwaSysDyn"""
        LTI1 = hybrid.LtiSysDyn(A=self.A1, B=self.B1,
                                Uset=self.Uset, domain=self.poly1,
                                time_semantics='sampled', timestep=.1)
        LTI2 = hybrid.LtiSysDyn(A=self.A2, B=self.B2,
                                Uset=self.Uset, domain=self.poly2,
                                time_semantics='sampled', timestep=.1)
        PWA = hybrid.PwaSysDyn(list_subsys=[LTI1, LTI2], domain=self.total_box,
                               time_semantics='discrete', overwrite_time=False)

    @raises(ValueError)
    def test_pwa_difftstep_from_subsys(self):
        """LtiSysDyn subsystems timesteps do not match that of PwaSysDyn"""
        LTI1 = hybrid.LtiSysDyn(A=self.A1, B=self.B1,
                                Uset=self.Uset, domain=self.poly1,
                                time_semantics='sampled', timestep=.1)
        LTI2 = hybrid.LtiSysDyn(A=self.A2, B=self.B2,
                                Uset=self.Uset, domain=self.poly2,
                                time_semantics='sampled', timestep=.1)
        PWA = hybrid.PwaSysDyn(list_subsys=[LTI1, LTI2], domain=self.total_box,
                               time_semantics='sampled', timestep=.2,
                               overwrite_time=False)

    @raises(ValueError)
    def test_pwa_invalid_semantics(self):
        LTI1 = hybrid.LtiSysDyn(A=self.A1, B=self.B1,
                                Uset=self.Uset, domain=self.poly1,
                                time_semantics='sampled', timestep=.1)
        LTI2 = hybrid.LtiSysDyn(A=self.A2, B=self.B2,
                                Uset=self.Uset, domain=self.poly2,
                                time_semantics='sampled', timestep=.1)
        PWA = hybrid.PwaSysDyn(list_subsys=[LTI1, LTI2], domain=self.total_box,
                               time_semantics='hello')

    @raises(ValueError)
    def test_disctime_errtstep(self):
        """Discrete time semantics yet given timestep"""
        LTI = hybrid.LtiSysDyn(A=self.A1, B=self.B1,
                               Uset=self.Uset, domain=self.poly1,
                               time_semantics='discrete', timestep=.1)


    @raises(ValueError)
    def test_lti_invalid_semantics(self):
        LTI = hybrid.LtiSysDyn(A=self.A1, B=self.B1,
                               Uset=self.Uset, domain=self.poly1,
                               time_semantics='hello', timestep=.1)

    @raises(ValueError)
    def test_nonpositive_timestep(self):
        LTI = hybrid.LtiSysDyn(A=self.A1, B=self.B1,
                               Uset=self.Uset, domain=self.poly1,
                               time_semantics='sampled', timestep=0)

    @raises(TypeError)
    def test_timestep_wrong_type(self):
        LTI = hybrid.LtiSysDyn(A=self.A1, B=self.B1,
                               Uset=self.Uset, domain=self.poly1,
                               time_semantics='sampled', timestep='.1')


class SwitchedSysDyn_test:
    def setUp(self):
        self.A1 = np.eye(2)
        self.A2 = np.array([[0, 1], [0, 0]])
        self.B1 = np.array([[0] ,[1]])
        self.B2 = np.array([[1], [0]])
        self.poly1 = pc.Polytope.from_box([[0, 1], [0, 1]])
        self.poly2 = pc.Polytope.from_box([[1, 2], [0, 1]])
        self.total_box = pc.Region(list_poly=[self.poly1, self.poly2])
        self.Uset = pc.Polytope.from_box([[0, 1]])
        self.env_labels = ('hi', 'hello')
        self.sys_labels = ('mode1',)
        self.disc_domain_size = (2, 1)
        self.LTI1 = hybrid.LtiSysDyn(A=self.A1, B=self.B1,
                                     Uset=self.Uset, domain=self.poly1,
                                     time_semantics='sampled', timestep=.1)
        self.LTI2 = hybrid.LtiSysDyn(A=self.A2, B=self.B2,
                                     Uset=self.Uset, domain=self.poly2,
                                     time_semantics='sampled', timestep=.1)
        self.LTI3 = hybrid.LtiSysDyn(A=self.A1, B=self.B1,
                                     Uset=self.Uset, domain=self.poly2,
                                     time_semantics='sampled', timestep=.1)
        self.LTI4 = hybrid.LtiSysDyn(A=self.A2, B=self.B2,
                                     Uset=self.Uset, domain=self.poly1,
                                     time_semantics='sampled', timestep=.1)
        self.PWA1 = hybrid.PwaSysDyn(list_subsys=[self.LTI1, self.LTI2],
                                     domain=self.total_box,
                                     time_semantics='sampled', timestep=.1)
        self.PWA2 = hybrid.PwaSysDyn(list_subsys=[self.LTI3, self.LTI4],
                                     domain=self.total_box,
                                     time_semantics='sampled', timestep=.1)
        self.dynamics1 = {(self.env_labels[0], self.sys_labels[0]): self.PWA1,
                          (self.env_labels[1], self.sys_labels[0]): self.PWA2}

    def tearDown(self):
        self.A1 = None
        self.A2 = None
        self.B1 = None
        self.B2 = None
        self.poly1 = None
        self.poly2 = None
        self.total_box = None
        self.Uset = None
        self.env_labels = None
        self.sys_labels = None
        self.disc_domain_size = None
        self.LTI1 = None
        self.LTI2 = None
        self.LTI3 = None
        self.LTI4 = None
        self.PWA1 = None
        self.PWA2 = None
        self.dynamics1 = None

    @raises(ValueError)
    def test_hybrid_difftstep_from_subsys(self):
        """LtiSysDyn subsystems timesteps do not match that of SwitchedSysDyn"""
        hybrid.SwitchedSysDyn(disc_domain_size=self.disc_domain_size,
                            dynamics=self.dynamics1, env_labels=self.env_labels,
                            disc_sys_labels=self.sys_labels,
                            time_semantics='hello', timestep=.1,
                            overwrite_time=True)

    @raises(ValueError)
    def test_hybrid_fail_check_time_consistency(self):
        # fail _check_time_consistency
        hybrid.SwitchedSysDyn(disc_domain_size=self.disc_domain_size,
                            dynamics=self.dynamics1, env_labels=self.env_labels,
                            disc_sys_labels=self.sys_labels,
                            time_semantics='sampled', timestep=.2,
                            overwrite_time=False)

    def test_correct_switched_construction(self):
        switched1 = hybrid.SwitchedSysDyn(disc_domain_size=self.disc_domain_size,
                                        dynamics=self.dynamics1,
                                        env_labels=self.env_labels,
                                        disc_sys_labels=self.sys_labels,
                                        time_semantics='sampled', timestep=.1,
                                        overwrite_time=True)
        switched2 = hybrid.SwitchedSysDyn(disc_domain_size=self.disc_domain_size,
                                        dynamics=self.dynamics1,
                                        env_labels=self.env_labels,
                                        disc_sys_labels=self.sys_labels,
                                        time_semantics='sampled', timestep=.1,
                                        overwrite_time=False)
        assert(switched1.time_semantics == 'sampled')
        assert(switched2.time_semantics == 'sampled')
        assert(switched1.timestep == .1)
        assert(switched2.timestep == .1)



# Copyright (c) 2013-2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Products between automata and transition systems"""
from __future__ import absolute_import
import logging
import warnings
from tulip.transys import transys
from tulip.transys import automata


logger = logging.getLogger(__name__)
_hl = 40 * '-'


class OnTheFlyProductAutomaton(automata.BuchiAutomaton):
    """Dynamically extends itself by adding successors.

    Note that performs on-the-fly BA * TS.
    The given TS can be explicit or on-the-fly,
    depending on what you pass to C{__init__}.

    The state space serves as the set of "visited" states.

    Note that this is equivalent to adding successors
    to both the queue and the set of visited states
    at the end of each iteration during a search,
    instead of adding each successor to the visited states
    when it is poped from the queue.
    """
    def __init__(self, ba, ts):
        self.ba = ba
        self.ts = ts
        super(OnTheFlyProductAutomaton, self).__init__()
        self.atomic_propositions |= ts.atomic_propositions
        self._add_initial()

    def _add_initial(self):
        ts = self.ts
        ba = self.ba
        s0s = set(ts.states.initial)
        q0s = set(ba.states.initial)

        logger.debug('\n' + _hl + '\n' +
                     ' Product BA construction:' +
                     '\n' + _hl + '\n')

        if not s0s:
            msg = (
                'Transition System has no initial states !\n'
                '=> Empty product system.\n'
                'Did you forget to define initial states ?'
            )
            warnings.warn(msg)

        for s0 in s0s:
            logger.debug('initial state:\t' + str(s0))

            for q0 in q0s:
                enabled_ba_trans = find_ba_succ(q0, s0, ts, ba)

                # q0 blocked ?
                if not enabled_ba_trans:
                    continue

                # which q next ?     (note: curq0 = q0)
                for (curq0, q, sublabels) in enabled_ba_trans:
                    new_sq0 = (s0, q)

                    self.states.add(new_sq0)
                    self.states.initial.add(new_sq0)

                    # accepting state ?
                    if q in ba.states.accepting:
                        self.states.accepting.add(new_sq0)

    def add_successors(self, s, q):
        """Add the successors of (s, q) to the state space.

        @param s: TS state

        @param q: BA state

        @return: those successors that are new states
        """
        sq = (s, q)
        ts = self.ts
        ba = self.ba

        logger.debug('Creating successors from'
                     ' product state:\t' + str(sq))

        # get next states
        next_ss = ts.states.post(s)
        next_sqs = set()
        for next_s in next_ss:
            enabled_ba_trans = find_ba_succ(q, next_s, ts, ba)

            if not enabled_ba_trans:
                continue

            (new_sqs, new_accepting) = find_prod_succ(
                sq, next_s, enabled_ba_trans,
                self, ba, ts
            )

            next_sqs.update(new_sqs)
            self.states.accepting |= new_accepting

        # new_sqs = {x for x in next_sqs if x not in self}

        logger.debug('next product states: ' + str(next_sqs))
        logger.debug('new unvisited product states: ' + str(new_sqs))

        return new_sqs

    def add_all_states(self):
        """Iterate L{add_successors} until all states are added.

        In other words until the state space
        reaches a fixed point.
        """
        Q = set(self.states)
        while Q:
            Qnew = set()
            for sq in Q:
                (s, q) = sq
                new = self.add_successors(s, q)
                Qnew.update(new)
            Q = Qnew


def ts_ba_sync_prod(transition_system, buchi_automaton):
    """Construct transition system for the synchronous product TS * BA.

    Def. 4.62, p.200 U{[BK08]
    <http://tulip-control.sourceforge.net/doc/bibliography.html#bk08>}

    Erratum
    =======
    note the erratum: P_{pers}(A) is ^_{q\in F} !q, verified from:
    http://www-i2.informatik.rwth-aachen.de/~katoen/errata.pdf

    See Also
    ========
    L{ba_ts_sync_prod}, L{sync_prod}

    @return: C{(product_ts, persistent_states)}, where:
        - C{product_ts} is the synchronous product TS * BA
        - C{persistent_states} are those in TS * BA which
            project on accepting states of BA.
    @rtype:
        - C{product_TS} is a L{transys.FiniteTransitionSystem}
        - C{persistent_states} is the set of states which project
            on accepting states of the Buchi Automaton BA.
    """
    # if not hasattr(transition_system, FiniteTransitionSystem):
    #    msg = 'transition_system not transys.FiniteTransitionSystem.\n'
    #    msg += 'Actual type passed: ' +str(type(transition_system) )
    #    raise TypeError(msg)

    if not hasattr(buchi_automaton, 'alphabet'):
        msg = 'transition_system not transys.BuchiAutomaton.\n'
        msg += 'Actual type passed: ' + str(type(buchi_automaton))
        raise TypeError(msg)

    if not buchi_automaton.atomic_proposition_based:
        msg = """Buchi automaton not stored as Atomic Proposition-based.
                synchronous product with Finite Transition System
                is not well-defined."""
        raise Exception(msg)

    fts = transition_system
    ba = buchi_automaton

    prodts_name = fts.name + '*' + ba.name
    prodts = transys.FiniteTransitionSystem()
    prodts.name = prodts_name

    prodts.atomic_propositions.add_from(ba.states())
    prodts.sys_actions.add_from(fts.actions)

    # construct initial states of product automaton
    s0s = set(fts.states.initial)
    q0s = set(ba.states.initial)

    accepting_states_preimage = set()

    logger.debug('\n' + _hl + '\n' +
                 ' Product TS construction:' +
                 '\n' + _hl + '\n')

    if not s0s:
        msg = (
            'Transition System has no initial states !\n'
            '=> Empty product system.\n'
            'Did you forget to define initial states ?')
        warnings.warn(msg)

    for s0 in s0s:
        logger.debug('initial state:\t' + str(s0))

        for q0 in q0s:
            enabled_ba_trans = find_ba_succ(q0, s0, fts, ba)

            # q0 blocked ?
            if not enabled_ba_trans:
                continue

            # which q next ?     (note: curq0 = q0)
            for (curq0, q, sublabels) in enabled_ba_trans:
                new_sq0 = (s0, q)
                prodts.states.add(new_sq0)
                prodts.states.initial.add(new_sq0)
                prodts.states[new_sq0]['ap'] = {q}

                # accepting state ?
                if q in ba.states.accepting:
                    accepting_states_preimage.add(new_sq0)

    # start visiting reachable in DFS or BFS way
    # (doesn't matter if we are going to store the result)
    queue = set(prodts.states.initial)
    visited = set()
    while queue:
        sq = queue.pop()
        visited.add(sq)
        (s, q) = sq

        logger.debug('Current product state:\t' + str(sq))

        # get next states
        next_ss = fts.states.post(s)
        next_sqs = set()
        for next_s in next_ss:
            enabled_ba_trans = find_ba_succ(q, next_s, fts, ba)

            if not enabled_ba_trans:
                continue

            (new_sqs, new_accepting) = find_prod_succ(
                sq, next_s, enabled_ba_trans,
                prodts, ba, fts)

            next_sqs.update(new_sqs)
            accepting_states_preimage.update(new_accepting)

        logger.debug('next product states: ' + str(next_sqs))
        # discard visited & push them to queue
        new_sqs = {x for x in next_sqs if x not in visited}
        logger.debug('new unvisited product states: ' + str(new_sqs))
        queue.update(new_sqs)

    return (prodts, accepting_states_preimage)


def find_ba_succ(prev_q, next_s, fts, ba):
    q = prev_q

    logger.debug('Next state:\t' + str(next_s))
    try:
        ap = fts.node[next_s]['ap']
    except:
        raise Exception(
            'No AP label for FTS state: ' + str(next_s) +
            '\n Did you forget labeing it ?')

    Sigma_dict = {'letter': ap}
    logger.debug("Next state's label:\t" + str(ap))

    enabled_ba_trans = ba.transitions.find(
        [q], with_attr_dict=Sigma_dict)
    enabled_ba_trans += ba.transitions.find(
        [q], letter={True})
    logger.debug('Enabled BA transitions:\n\t' +
                 str(enabled_ba_trans))

    if not enabled_ba_trans:
        logger.debug('No enabled BA transitions at: ' + str(q))

    logger.debug('---\n')

    return enabled_ba_trans


def find_prod_succ(prev_sq, next_s, enabled_ba_trans, product, ba, fts):
    (s, q) = prev_sq

    new_accepting = set()
    next_sqs = set()
    for (curq, next_q, sublabels) in enabled_ba_trans:
        assert(curq == q)

        new_sq = (next_s, next_q)

        if new_sq not in product:
            next_sqs.add(new_sq)
            product.states.add(new_sq)

            logger.debug('Adding state:\t' + str(new_sq))

        if hasattr(product, 'actions'):
            product.states[new_sq]['ap'] = {next_q}

        # accepting state ?
        if next_q in ba.states.accepting:
            new_accepting.add(new_sq)
            logger.debug(str(new_sq) +
                         ' contains an accepting state.')

        logger.debug('Adding transitions:\t' +
                     str(prev_sq) + '--->' + str(new_sq))

        # is fts transition labeled with an action ?
        enabled_ts_trans = fts.transitions.find(
            [s], to_states=[next_s],
            with_attr_dict=None)
        for (from_s, to_s, sublabel_values) in enabled_ts_trans:
            assert(from_s == s)
            assert(to_s == next_s)

            logger.debug('Sublabel value:\n\t' +
                         str(sublabel_values))

            # labeled transition ?
            if hasattr(product, 'alphabet'):
                product.transitions.add(
                    prev_sq, new_sq,
                    letter=fts.states[to_s]['ap'])
            elif hasattr(product, 'actions'):
                if not sublabel_values:
                    product.transitions.add(prev_sq, new_sq)
                else:
                    product.transitions.add(
                        prev_sq, new_sq,
                        actions=sublabel_values['actions'])
    return (next_sqs, new_accepting)


def ba_ts_sync_prod(buchi_automaton, transition_system):
    """Construct Buchi Automaton equal to synchronous product TS x NBA.

    See Also
    ========
    L{ts_ba_sync_prod}, L{sync_prod}

    @return: C{prod_ba}, the product L{BuchiAutomaton}.
    """
    logger.debug('\n' + _hl + '\n'
                 'Product: BA * TS' +
                 '\n' + _hl + '\n')

    (prod_ts, persistent) = ts_ba_sync_prod(
        transition_system, buchi_automaton)

    prod_name = buchi_automaton.name + '*' + transition_system.name

    prod_ba = automata.BuchiAutomaton()
    prod_ba.name = prod_name

    # copy S, S0, from prod_TS-> prod_BA
    prod_ba.states.add_from(prod_ts.states())
    prod_ba.states.initial |= set(prod_ts.states.initial)

    # accepting states = persistent set
    prod_ba.states.accepting |= persistent

    # copy edges, translating transitions,
    # i.e., changing transition labels
    if not buchi_automaton.atomic_proposition_based:
        msg = (
            'Buchi Automaton must be Atomic Proposition-based,'
            ' otherwise the synchronous product is not well-defined.')
        raise Exception(msg)

    # direct access, not the inefficient
    #   prod_ba.alphabet.add_from(buchi_automaton.alphabet() ),
    # which would generate a combinatorially large alphabet
    prod_ba.alphabet.math_set |= buchi_automaton.alphabet.math_set

    for (from_state, to_state) in prod_ts.transitions():
        # prject prod_TS state to TS state
        ts_to_state = to_state[0]
        msg = (
            'prod_TS: to_state =\n\t' + str(to_state) + '\n'
            'TS: ts_to_state =\n\t' + str(ts_to_state))
        logger.debug(msg)

        state_label_pairs = transition_system.states.find(ts_to_state)
        (ts_to_state_, transition_label_dict) = state_label_pairs[0]
        transition_label_value = transition_label_dict['ap']
        prod_ba.transitions.add(
            from_state, to_state, letter=transition_label_value)
    return prod_ba



import numpy


def matrix2str(A):
    """Convert a matrix A into a string

    @param A: a numpy.array matrix
    @rtype string
    """
    s = ""
    for x in numpy.nditer(A, order='F'):
        s = s + str(x) + ","

    return s


def polytope2str(p, name):
    """Convert a polytope into C code

    @param p: a polytope.polytope.Polytope
    @param name: the name of the polytope in C code
    @rtype string
    """
    k = p.A.shape[0]
    l = p.A.shape[1]
    # pik=k
    s = "idxint "+name+"k = "+str(k)+";\n"
    # pil=l
    s = s+"idxint "+name+"l = "+str(l)+";\n"
    # piA = [A11,A21,...,Ak1,A12,...,Ak2,...,A1l,...,Akl];
    s = s+"pfloat "+name+"A[] = {"+matrix2str(-1*p.A)+"};\n"
    # pib = [b1,b2,...,bk];
    s = s+"pfloat "+name+"b[] = {"+matrix2str(-1*p.b)+"};\n"
    # picenter = [p.chebXc1,...,p.chebXcl];
    s = s+"pfloat "+name+"center[] = {"+matrix2str(p.chebXc)+"};\n"
    return s



# Copyright (c) 2014, 2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Interface to the slugs implementation of GR(1) synthesis.

Relevant links:
  - U{slugs<https://github.com/LTLMoP/slugs>}
"""
from __future__ import absolute_import
import logging
import json
import os
import subprocess
import tempfile
import networkx as nx
from tulip.spec import GRSpec, translate


# If this path begins with '/', then it is considered to be absolute.
# Otherwise, it is relative to the path of the `slugs` executable.
SLUGS_COMPILER_PATH = '../tools/StructuredSlugsParser/compiler.py'

BDD_FILE = 'strategy_bdd.txt'
logger = logging.getLogger(__name__)


def check_realizable(spec):
    """Decide realizability of specification.

    Consult the documentation of L{synthesize} about parameters.

    @return: True if realizable, False if not, or an error occurs.
    """
    if isinstance(spec, GRSpec):
        assert not spec.moore
        assert not spec.plus_one
        struct = translate(spec, 'slugs')
    else:
        struct = spec
    with tempfile.NamedTemporaryFile(delete=False) as fin:
        fin.write(struct)
    realizable, out = _call_slugs(fin.name, synth=False)
    return realizable


def synthesize(spec, symbolic=False):
    """Return strategy satisfying the specification C{spec}.

    @type spec: L{GRSpec} or C{str} in structured slugs syntax.
    @return: If realizable return synthesized strategy, otherwise C{None}.
    @rtype: C{networkx.DiGraph}
    """
    if isinstance(spec, GRSpec):
        assert not spec.moore
        assert not spec.plus_one
        struct = translate(spec, 'slugs')
    else:
        struct = spec
    with tempfile.NamedTemporaryFile(delete=False) as fin:
        fin.write(struct)
    realizable, out = _call_slugs(fin.name, synth=True, symbolic=symbolic)
    if not realizable:
        return None
    os.unlink(fin.name)
    # collect int vars
    vrs = dict(spec.sys_vars)
    vrs.update(spec.env_vars)
    dout = json.loads(out)
    g = nx.DiGraph()
    dvars = dout['variables']
    for stru, d in dout['nodes'].iteritems():
        u = int(stru)
        state = dict(zip(dvars, d['state']))
        g.add_node(u, state=state)
        for v in d['trans']:
            g.add_edge(u, v)
    h = nx.DiGraph()
    for u, d in g.nodes_iter(data=True):
        bit_state = d['state']
        int_state = _bitfields_to_ints(bit_state, vrs)
        h.add_node(u, state=int_state)
    for u, v in g.edges_iter():
        h.add_edge(u, v)
    logger.debug(
        ('loaded strategy with vertices:\n  {v}\n'
         'and edges:\n {e}\n').format(
            v='\n  '.join(str(x) for x in h.nodes(data=True)),
            e=h.edges()))
    return h


def _bitfields_to_ints(bit_state, vrs):
    """Convert bitfield representation to integers.

    @type line: C{str}
    @type vrs: C{dict}
    """
    int_state = dict()
    for var, dom in vrs.iteritems():
        if dom == 'boolean':
            int_state[var] = bit_state[var]
            continue
        bitnames = ['{var}@{i}'.format(var=var, i=i)
                    for i in xrange(dom[1].bit_length())]
        bitnames[0] = '{var}@0.{min}.{max}'.format(
            var=var, min=dom[0], max=dom[1])
        bitvalues = [bit_state[b] for b in bitnames]
        # little-endian
        val = int(''.join(str(b) for b in reversed(bitvalues)), 2)
        int_state[var] = val
    return int_state


def _call_slugs(filename, synth=True, symbolic=True, slugs_compiler_path=None):
    """Call `slugs` and return results.

    slugs_compiler_path is the path to the slugsin converter format.
    If None (default), then use the path as in the module-level
    identifier SLUGS_COMPILER_PATH.  If this path begins with '/',
    then it is considered to be absolute.  Otherwise, it is relative
    to the path of the `slugs` executable.
    """
    if slugs_compiler_path is None:
        slugs_compiler_path = SLUGS_COMPILER_PATH

    slugs_path = None
    for exe_path in os.environ['PATH'].split(':'):
        if os.path.exists(os.path.join(exe_path, 'slugs')):
            slugs_path = os.path.join(exe_path, 'slugs')
            break

    if slugs_path is None:
        raise Exception('slugs not found in path.')

    if not os.path.isabs(slugs_compiler_path):
        slugs_compiler_path = os.path.abspath(
            os.path.join(os.path.dirname(slugs_path),
                         slugs_compiler_path))

    if not os.path.exists(slugs_compiler_path):
        raise Exception('slugs/compiler.py not found.')

    with tempfile.NamedTemporaryFile(delete=False) as slugs_infile:
        subprocess.check_call([slugs_compiler_path, filename],
                              stdout=slugs_infile,
                              stderr=subprocess.STDOUT)

    options = [slugs_path, slugs_infile.name]
    if synth:
        if symbolic:
            options.extend(['--symbolicStrategy', BDD_FILE])
        else:
            options.append('--explicitStrategy')
            options.append('--jsonOutput')
    else:
        # As of commit ad0cf12c14131fc6a20fe29edfe04d9aefd7c6d4
        # (tip of master branch of
        #  https://github.com/LTLMoP/slugs.git at the time of writing),
        # Slugs seems to default to checking realizability. Trying to
        # provide `--onlyRealizability` leads to error message from
        # `slugs`: "Error: Parameter '--onlyRealizability' is unknown."
        pass
    logger.debug('Calling: ' + ' '.join(options))
    try:
        p = subprocess.Popen(
            options,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
    except OSError as e:
        if e.errno == os.errno.ENOENT:
            raise Exception('slugs not found in path.')
        else:
            raise
    out, err = p.communicate()
    msg = (
        '\n slugs return code: {c}\n\n'.format(c=p.returncode) +
        '\n slugs stderr: {c}\n\n'.format(c=err) +
        '\n slugs stdout:\n\n {out}\n\n'.format(out=out))
    logger.debug(msg)
    # error ?
    if p.returncode != 0:
        raise Exception(msg)
    realizable = 'Specification is realizable' in err
    # check sanity
    if not realizable:
        assert 'Specification is unrealizable' in err
    return realizable, out



#!/usr/bin/env python
"""Tests for GR(1) specification formulas module."""
import logging
logging.basicConfig(level=logging.ERROR)
logging.getLogger('ltl_parser_log').setLevel(logging.WARNING)
import nose.tools as nt
from tulip.spec.form import LTL, GRSpec, replace_dependent_vars


class LTL_test:
    def setUp(self):
        self.f = LTL("[](p -> <>q)", input_variables={"p": "boolean"},
                     output_variables={"q": "boolean"})

    def tearDown(self):
        self.f = None

    def test_loads_dumps_id(self):
        # Dump/load identity test: Dumping the result from loading a
        # dump should be identical to the original dump.
        assert self.f.dumps() == LTL.loads(self.f.dumps()).dumps()


def GR1specs_equal(s1, s2):
    """Return True if s1 and s2 are *roughly* syntactically equal.

    This function seems to be of little or no use outside this test
    module because of its fragility.
    """
    if s1 is None or s2 is None:
        raise TypeError
    for s in [s1, s2]:
        if hasattr(s.env_init, "sort"):
            s.env_init.sort()
        if hasattr(s.env_safety, "sort"):
            s.env_safety.sort()
        if hasattr(s.env_prog, "sort"):
            s.env_prog.sort()
        if hasattr(s.sys_init, "sort"):
            s.sys_init.sort()
        if hasattr(s.sys_safety, "sort"):
            s.sys_safety.sort()
        if hasattr(s.sys_prog, "sort"):
            s.sys_prog.sort()
    if s1.env_vars != s2.env_vars or s1.sys_vars != s2.sys_vars:
        return False
    if (
        s1.env_init != s2.env_init or
        s1.env_safety != s2.env_safety or
        s1.env_prog != s2.env_prog
    ):
        return False
    if (
        s1.sys_init != s2.sys_init or
        s1.sys_safety != s2.sys_safety or
        s1.sys_prog != s2.sys_prog
    ):
        return False
    return True


class GRSpec_test:
    def setUp(self):
        self.f = GRSpec(env_vars={"x"}, sys_vars={"y"},
                        env_init=["x"], sys_safety=["y"],
                        env_prog=["!x", "x"], sys_prog=["y&&!x"])
        self.triv = GRSpec(env_vars=["x"], sys_vars=["y"],
                           env_init=["x && !x"])
        self.empty = GRSpec()

    def tearDown(self):
        self.f = None

    # def test_sym_to_prop(self):
    #    original_env_vars = copy.copy(self.f.env_vars)
    #    original_sys_vars = copy.copy(self.f.sys_vars)
    #    self.f.sym_to_prop({"x":"bar", "y":"uber||cat"})
    #    assert self.f.env_vars == original_env_vars and self.f.sys_vars == original_sys_vars
    #    assert self.f.env_prog == ["!(bar)", "(bar)"] and self.f.sys_prog == ["(uber||cat)&&!(bar)"]

    def test_or(self):
        g = GRSpec(env_vars={"z"}, env_prog=["!z"])
        h = self.f | g
        assert len(h.env_vars) == 2 and 'z' in h.env_vars
        assert (
            len(h.env_prog) == len(self.f.env_prog) + 1 and
            '!z' in h.env_prog)

        # Domain mismatch on system variable y
        g.sys_vars = {"y": (0, 5)}
        nt.assert_raises(ValueError, self.f.__or__, g)

        # Domain mismatch on environment variable x
        g.sys_vars = dict()
        g.env_vars["x"] = (0, 3)
        nt.assert_raises(ValueError, self.f.__or__, g)

    def test_to_canon(self):
        # Fragile!
        assert (self.f.to_canon() ==
                "((x) && []<>(!x) && []<>(x)) -> ([](y) && []<>(y&&!x))")
        # N.B., for self.triv, to_canon() returns a formula missing
        # the assumption part not because it detected that the
        # assumption is false, but rather the guarantee is empty (and
        # thus interpreted as being "True").
        assert self.triv.to_canon() == "True"
        assert self.empty.to_canon() == "True"

    def test_init(self):
        assert len(self.f.env_vars) == 1 and len(self.f.sys_vars) == 1
        assert (self.f.env_vars["x"] == "boolean" and
                self.f.sys_vars["y"] == "boolean")


def test_str_to_int():
    x = "a' = \"hehe\""
    s = GRSpec(sys_vars={'a': ['hehe', 'haha']},
               sys_safety=[x])
    s.str_to_int()
    assert x in s._ast
    assert x in s._bool_int
    print(s._bool_int[x])
    assert s._bool_int[x] == "( ( X a ) = 0 )"


def test_compile_init():
    env_vars = {'x': (0, 0), 'y': (0, 0), 'z': (0, 1)}
    sys_vars = {'w': (0, 0)}
    env_init = ['((((y = 0))) & (x = 0))']
    sys_init = ['((w = 0))']
    spc = GRSpec(
        env_vars=env_vars, sys_vars=sys_vars,
        env_init=env_init, sys_init=sys_init)
    code = spc.compile_init(no_str=True)
    d = dict(x=0, y=0, z=0, w=0)
    assert eval(code, d)
    d = dict(x=0, y=1, z=1, w=0)
    assert eval(code, d)
    d = dict(x=0, y=0, z=0, w=1)
    assert not eval(code, d)


def test_replace_dependent_vars():
    sys_vars = {'a': 'boolean', 'locA': (0, 4)}
    sys_safe = ['!a', 'a & (locA = 3)']
    spc = GRSpec(sys_vars=sys_vars, sys_safety=sys_safe)
    bool2form = {'a': '(locA = 0) | (locA = 2)', 'b': '(locA = 1)'}
    replace_dependent_vars(spc, bool2form)
    correct_result = (
        '[](( ! ( ( locA = 0 ) | ( locA = 2 ) ) )) && '
        '[](( ( ( locA = 0 ) | ( locA = 2 ) ) & ( locA = 3 ) ))')
    assert str(spc) == correct_result



# Copyright (c) 2011-2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""
Check Linear Discrete-Time-Invariant System reachability between polytopes

Primary functions:
    - L{solve_feasible}
    - L{createLM}
    - L{get_max_extreme}

See Also
========
L{find_controller}
"""
import logging
logger = logging.getLogger(__name__)

from collections import Iterable

import numpy as np
import polytope as pc

def is_feasible(
    from_region, to_region, sys, N,
    closed_loop=True,
    use_all_horizon=False,
    trans_set=None
):
    """Return True if to_region is reachable from_region.

    For details see solve_feasible.
    """
    S0 = solve_feasible(
        from_region, to_region, sys, N,
        closed_loop, use_all_horizon,
        trans_set
    )
    return from_region <= S0

def solve_feasible(
    P1, P2, ssys, N=1, closed_loop=True,
    use_all_horizon=False, trans_set=None, max_num_poly=5
):
    r"""Compute S0 \subseteq trans_set from which P2 is N-reachable.

    N-reachable = reachable in horizon C{N}.
    The system dynamics are C{ssys}.
    The closed-loop algorithm solves for one step at a time,
    which keeps the dimension of the polytopes down.

    Time semantics:

    - C{use_all_horizon = False}: fixed sampling period of
      discrete-valued environment variables.
      Reachability in exactly C{N} steps.

    - C{use_all_horizon = True}: sampling period that varies and
      is chosen by the system, depending on how many steps are
      taken during each trajectory from C{P1} to C{P2}.
      Reachability in C{1..N} steps, with an under-approximation
      of the attractor set.

      If the system dynamics do not allow staying at the same state,
      then disconnected polytopes can arise, possibly causing issues
      in the computation. Consider decreasing the sampling period
      used for discretizing the associated continuous-time dynamical system.

    @type P1: C{Polytope} or C{Region}
    @type P2: C{Polytope} or C{Region}
    @type ssys: L{LtiSysDyn}
    @param N: horizon length
    @param closed_loop: If C{True}, then take 1 step at a time.
        This keeps down polytope dimension and
        handles disturbances better.
    @type closed_loop: bool

    @param use_all_horizon: Used for closed loop algorithm.

        - If C{True}, then check for reachability in C{< N} steps,

        - otherwise, in exactly C{N} steps.

    @type use_all_horizon: bool

    @param trans_set: If specified,
        then force transitions to be in this set.
        Otherwise, P1 is used.

    @return: states from which P2 is reachable
    @rtype: C{Polytope} or C{Region}
    """
    if closed_loop:
        if use_all_horizon:
            return _underapproximate_attractor(
                P1, P2, ssys, N, trans_set=trans_set)
        else:
            return _solve_closed_loop_fixed_horizon(
                P1, P2, ssys, N, trans_set=trans_set)
    else:
        if use_all_horizon:
            raise ValueError(
                '`use_all_horizon = True` has no effect if '
                '`closed_loop = False`')
        return solve_open_loop(
            P1, P2, ssys, N,
            trans_set=trans_set,
            max_num_poly=max_num_poly
        )


def _solve_closed_loop_fixed_horizon(
        P1, P2, ssys, N, trans_set=None):
    """Under-approximate states in P1 that can reach P2 in N > 0 steps.

    If intermediate polytopes are convex,
    then the result is exact and not an under-approximation.

    @type P1: C{Polytope} or C{Region}
    @type P2: C{Polytope} or C{Region}

    @param ssys: system dynamics

    @param N: horizon length
    @type N: int > 0

    @param trans_set: If provided,
        then intermediate steps are allowed
        to be in trans_set.

        Otherwise, P1 is used.
    """
    assert N > 0, N
    p1 = P1.copy()  # initial set
    p2 = P2.copy()  # terminal set
    if trans_set is None:
        pinit = p1
    else:
        pinit = trans_set
    # backwards in time
    for i in xrange(N, 0, -1):
        # first step from P1
        if i == 1:
            pinit = p1
        p2 = solve_open_loop(pinit, p2, ssys, 1, trans_set)
        p2 = pc.reduce(p2)
        if not pc.is_fulldim(p2):
            return pc.Polytope()
    return p2


def _solve_closed_loop_bounded_horizon(
        P1, P2, ssys, N, trans_set=None):
    """Under-approximate states in P1 that can reach P2 in <= N steps.

    See docstring of function `_solve_closed_loop_fixed_horizon`
    for details.
    """
    _print_horizon_warning()
    p1 = P1.copy()  # initial set
    p2 = P2.copy()  # terminal set
    if trans_set is None:
        pinit = p1
    else:
        pinit = trans_set
    # backwards in time
    s = pc.Region()
    for i in xrange(N, 0, -1):
        # first step from P1
        if i == 1:
            pinit = p1
        p2 = solve_open_loop(pinit, p2, ssys, 1, trans_set)
        p2 = pc.reduce(p2)
        # running union
        s = s.union(p2, check_convex=True)
        s = pc.reduce(s)
        # empty target polytope ?
        if not pc.is_fulldim(p2):
            break
    if not pc.is_fulldim(s):
        return pc.Polytope()
    s = pc.reduce(s)
    return s


def _underapproximate_attractor(
        P1, P2, ssys, N, trans_set=None):
    """Under-approximate N-step attractor of polytope P2, with N > 0.

    See docstring of function `_solve_closed_loop_fixed_horizon`
    for details.
    """
    assert N > 0, N
    _print_horizon_warning()
    p1 = P1.copy()  # initial set
    p2 = P2.copy()  # terminal set
    if trans_set is None:
        pinit = p1
    else:
        pinit = trans_set
    # backwards in time
    for i in xrange(N, 0, -1):
        # first step from P1
        if i == 1:
            pinit = p1
        r = solve_open_loop(pinit, p2, ssys, 1, trans_set)
        p2 = p2.union(r, check_convex=True)
        p2 = pc.reduce(p2)
        # empty target polytope ?
        if not pc.is_fulldim(p2):
            return pc.Polytope()
    return r


def _print_horizon_warning():
    print('WARNING: different timing semantics and assumptions '
          'from the case of fixed horizon. '
          'Also, depending on dynamics, disconnected polytopes '
          'can arise, which may cause issues in '
          'the `polytope` package.')


def solve_open_loop(
    P1, P2, ssys, N,
    trans_set=None, max_num_poly=5
):
    r1 = P1.copy() # Initial set
    r2 = P2.copy() # Terminal set

    # use the max_num_poly largest volumes for reachability
    r1 = volumes_for_reachability(r1, max_num_poly)
    r2 = volumes_for_reachability(r2, max_num_poly)

    if len(r1) > 0:
        start_polys = r1
    else:
        start_polys = [r1]

    if len(r2) > 0:
        target_polys = r2
    else:
        target_polys = [r2]

    # union of s0 over all polytope combinations
    s0 = pc.Polytope()
    for p1 in start_polys:
        for p2 in target_polys:
            cur_s0 = poly_to_poly(p1, p2, ssys, N, trans_set)
            s0 = s0.union(cur_s0, check_convex=True)

    return s0

def poly_to_poly(p1, p2, ssys, N, trans_set=None):
    """Compute s0 for open-loop polytope to polytope N-reachability.
    """
    p1 = p1.copy()
    p2 = p2.copy()

    if trans_set is None:
        trans_set = p1

    # stack polytope constraints
    L, M = createLM(ssys, N, p1, trans_set, p2)
    s0 = pc.Polytope(L, M)
    s0 = pc.reduce(s0)

    # Project polytope s0 onto lower dim
    n = np.shape(ssys.A)[1]
    dims = range(1, n+1)

    s0 = s0.project(dims)

    return pc.reduce(s0)

def volumes_for_reachability(part, max_num_poly):
    if len(part) <= max_num_poly:
        return part

    vol_list = np.zeros(len(part) )
    for i in xrange(len(part) ):
        vol_list[i] = part[i].volume

    ind = np.argsort(-vol_list)
    temp = []
    for i in ind[range(max_num_poly) ]:
        temp.append(part[i] )

    part = pc.Region(temp, [])
    return part

def createLM(ssys, N, list_P, Pk=None, PN=None, disturbance_ind=None):
    """Compute the components of the polytope::

        L [x(0)' u(0)' ... u(N-1)']' <= M

    which stacks the following constraints:

      - x(t+1) = A x(t) + B u(t) + E d(t)
      - [u(k); x(k)] \in ssys.Uset for all k

    If list_P is a C{Polytope}:

      - x(0) \in list_P if list_P
      - x(k) \in Pk for k= 1,2, .. N-1
      - x(N) \in PN

    If list_P is a list of polytopes:

      - x(k) \in list_P[k] for k= 0, 1 ... N

    The returned polytope describes the intersection of the polytopes
    for all possible

    @param ssys: system dynamics
    @type ssys: L{LtiSysDyn}

    @param N: horizon length

    @type list_P: list of Polytopes or C{Polytope}
    @type Pk: C{Polytope}
    @type PN: C{Polytope}

    @param disturbance_ind: list indicating which k's
        that disturbance should be taken into account.
        Default is [1,2, ... N]
    """
    if not isinstance(list_P, Iterable):
        list_P = [list_P] +(N-1) *[Pk] +[PN]

    if disturbance_ind is None:
        disturbance_ind = range(1,N+1)

    A = ssys.A
    B = ssys.B
    E = ssys.E
    K = ssys.K

    D = ssys.Wset
    PU = ssys.Uset

    n = A.shape[1]  # State space dimension
    m = B.shape[1]  # Input space dimension
    p = E.shape[1]  # Disturbance space dimension

    # non-zero disturbance matrix E ?
    if not np.all(E==0):
        if not pc.is_fulldim(D):
            E = np.zeros(K.shape)

    list_len = np.array([P.A.shape[0] for P in list_P])
    sumlen = np.sum(list_len)

    LUn = np.shape(PU.A)[0]

    Lk = np.zeros([sumlen, n+N*m])
    LU = np.zeros([LUn*N, n+N*m])

    Mk = np.zeros([sumlen, 1])
    MU = np.tile(PU.b.reshape(PU.b.size, 1), (N, 1))

    Gk = np.zeros([sumlen, p*N])
    GU = np.zeros([LUn*N, p*N])

    K_hat = np.tile(K, (N, 1))

    B_diag = B
    E_diag = E
    for i in xrange(N-1):
        B_diag = _block_diag2(B_diag, B)
        E_diag = _block_diag2(E_diag, E)

    A_n = np.eye(n)
    A_k = np.zeros([n, n*N])

    sum_vert = 0
    for i in xrange(N+1):
        Li = list_P[i]

        if not isinstance(Li, pc.Polytope):
            logger.warn('createLM: Li of type: ' +str(type(Li) ) )

        ######### FOR M #########
        idx = range(sum_vert, sum_vert + Li.A.shape[0])
        Mk[idx, :] = Li.b.reshape(Li.b.size,1) - \
                     Li.A.dot(A_k).dot(K_hat)

        ######### FOR G #########
        if i in disturbance_ind:
            idx = np.ix_(
                range(sum_vert, sum_vert + Li.A.shape[0]),
                range(Gk.shape[1])
            )
            Gk[idx] = Li.A.dot(A_k).dot(E_diag)

            if (PU.A.shape[1] == m+n) and (i < N):
                A_k_E_diag = A_k.dot(E_diag)
                d_mult = np.vstack([np.zeros([m, p*N]), A_k_E_diag])

                idx = np.ix_(range(LUn*i, LUn*(i+1)), range(p*N))
                GU[idx] = PU.A.dot(d_mult)

        ######### FOR L #########
        AB_line = np.hstack([A_n, A_k.dot(B_diag)])

        idx = np.ix_(
            range(sum_vert, sum_vert + Li.A.shape[0]),
            range(0,Lk.shape[1])
        )
        Lk[idx] = Li.A.dot(AB_line)

        if i >= N:
            continue

        if PU.A.shape[1] == m:
            idx = np.ix_(
                range(i*LUn, (i+1)*LUn),
                range(n + m*i, n + m*(i+1))
            )
            LU[idx] = PU.A
        elif PU.A.shape[1] == m+n:
            uk_line = np.zeros([m, n + m*N])

            idx = np.ix_(range(m), range(n+m*i, n+m*(i+1)))
            uk_line[idx] = np.eye(m)

            A_mult = np.vstack([uk_line, AB_line])

            b_mult = np.zeros([m+n, 1])
            b_mult[range(m, m+n), :] = A_k.dot(K_hat)

            idx = np.ix_(
                range(i*LUn, (i+1)*LUn),
                range(n+m*N)
            )
            LU[idx] = PU.A.dot(A_mult)

            MU[range(i*LUn, (i+1)*LUn), :] -= PU.A.dot(b_mult)

        ####### Iterate #########
        sum_vert += Li.A.shape[0]
        A_n = A.dot(A_n)
        A_k = A.dot(A_k)

        idx = np.ix_(range(n), range(i*n, (i+1)*n))
        A_k[idx] = np.eye(n)

    # Get disturbance sets
    if not np.all(Gk==0):
        G = np.vstack([Gk, GU])
        D_hat = get_max_extreme(G, D, N)
    else:
        D_hat = np.zeros([sumlen + LUn*N, 1])

    # Put together matrices L, M
    L = np.vstack([Lk, LU])
    M = np.vstack([Mk, MU]) - D_hat

    msg = 'Computed S0 polytope: L x <= M, where:\n\t L = \n'
    msg += str(L) +'\n\t M = \n' + str(M) +'\n'
    logger.debug(msg)

    return L,M

def get_max_extreme(G,D,N):
    """Calculate the array d_hat such that::

        d_hat = max(G*DN_extreme),

    where DN_extreme are the vertices of the set D^N.

    This is used to describe the polytope::

        L*x <= M - G*d_hat.

    Calculating d_hat is equivalen to taking the intersection
    of the polytopes::

        L*x <= M - G*d_i

    for every possible d_i in the set of extreme points to D^N.

    @param G: The matrix to maximize with respect to
    @param D: Polytope describing the disturbance set
    @param N: Horizon length

    @return: d_hat: Array describing the maximum possible
        effect from the disturbance
    """
    D_extreme = pc.extreme(D)
    nv = D_extreme.shape[0]
    dim = D_extreme.shape[1]
    DN_extreme = np.zeros([dim*N, nv**N])

    for i in xrange(nv**N):
        # Last N digits are indices we want!
        ind = np.base_repr(i, base=nv, padding=N)
        for j in xrange(N):
            DN_extreme[range(j*dim,(j+1)*dim),i] = D_extreme[int(ind[-j-1]),:]

    d_hat = np.amax(np.dot(G,DN_extreme), axis=1)
    return d_hat.reshape(d_hat.size,1)

def _block_diag2(A,B):
    """Like block_diag() in scipy.linalg, but restricted to 2 inputs.

    Old versions of the linear algebra package in SciPy (i.e.,
    scipy.linalg) do not have a block_diag() function.  Providing
    _block_diag2() here until most folks are using sufficiently
    up-to-date SciPy installations improves portability.
    """
    if len(A.shape) == 1:  # Cast 1d array into matrix
        A = np.array([A])
    if len(B.shape) == 1:
        B = np.array([B])
    C = np.zeros((A.shape[0]+B.shape[0], A.shape[1]+B.shape[1]))
    C[:A.shape[0], :A.shape[1]] = A.copy()
    C[A.shape[0]:, A.shape[1]:] = B.copy()
    return C



# Copyright (c) 2014-2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Code generation and exporting of controllers from TuLiP

Routines in this module are cross-cutting in the sense that they
concern multiple aspects of solutions created by TuLiP and accordingly
should not be placed under a specific subpackage, like tulip.transys.
"""
from itertools import chain, repeat
import time


def write_python_case(filename, *args, **kwargs):
    """Convenience wrapper for writing output of python_case to file.

    @type  filename: str
    @param filename: Name of file in which to place the code generated
        by L{python_case}.
    """
    with open(filename, 'w') as f:
        f.write(python_case(*args, **kwargs))


def python_case(M, classname="TulipStrategy", start='Sinit'):
    """Export MealyMachine as Python class based on flat if-else block.

    Usage documentation for the generated code is included in the output.
    Consult the docstrings of the class and move() method.

    @type M: L{MealyMachine}
    @type classname: C{str}
    @param start: initial node in C{M}

    @rtype: str
    @return: The returned string is valid Python code and can, for
        example, be:
          - saved directly into a ".*.py" file, or
          - passed to "exec".
    """
    tab = 4 * ' '
    node_to_int = dict([(s, i) for i, s in enumerate(M)])
    input_vars = [input_var for input_var in M.inputs] if M.inputs else []
    input_args = ', '.join(input_vars)
    input_args_str = "'"+"', '".join(input_vars)+"'"
    code = (
        'class {classname}(object):\n'
        '{t}"""Mealy transducer.\n'
        '\n'
        '{t}Internal states are integers, the current state\n'
        '{t}is stored in the attribute "state".\n'
        '{t}To take a transition, call method "move".\n'
        '\n'
        '{t}The names of input variables are stored in the\n'
        '{t}attribute "input_vars".\n'
        '\n'
        '{t}Automatically generated by tulip.dumpsmach on {date}\n'
        '{t}To learn more about TuLiP, visit http://tulip-control.org\n'
        '{t}"""\n'
        '{t}def __init__(self):\n'
        '{t2}self.state = {sinit}\n'
        '{t2}self.input_vars = [{input_args_str}]\n'
        '\n'
        '{t}def move(self, {input_args}):\n'
        '{t2}"""Given inputs, take move and return outputs.\n'
        '\n'
        '{t2}@rtype: dict\n'
        '{t2}@return: dictionary with keys of the output variable names:\n'
        '{t2}    {outputs}\n'
        '{t2}"""\n'
        '{t2}output = dict()\n'
        ).format(
            classname=classname,
            t=tab,
            t2=2*tab,
            date=time.strftime('%Y-%m-%d %H:%M:%S UTC', time.gmtime()),
            sinit=node_to_int[start],
            input_args_str=input_args_str,
            input_args=input_args,
            outputs=[str(v) for v in M.outputs])
    # cached generator
    ifs = lambda: chain(['if'], repeat('elif'))
    proj = lambda d, keys: ((k, d[k]) for k in d if k in keys)
    proj = lambda d, keys: ((k, "'"+d[k]+"'" if isinstance(d[k], str) else d[k])
                            for k in d if k in keys)
    # generate selection statements
    c = list()
    for u, ifu in zip(M, ifs()):
        edges = list()
        for (_, w, d), ifw in zip(M.edges_iter(u, data=True), ifs()):
            if M.inputs:
                guard = ' and '.join(
                    '({k} == {v})'.format(k=k, v=v)
                    for k, v in proj(d, M.inputs))
            else:
                guard = 'True'
            outputs = ''.join(
                '{t4}output["{k}"] = {v}\n'.format(k=k, v=v, t4=4*tab)
                for k, v in proj(d, M.outputs))
            edges.append((
                '{t3}{ifw} {guard}:\n'
                '{t4}self.state = {target_id}\n'
                '\n'
                '{outputs}').format(
                    t3=3*tab,
                    t4=4*tab,
                    ifw=ifw,
                    guard=guard,
                    target_id=node_to_int[w],
                    outputs=outputs))
        # handle invalid inputs or dead-end
        if edges and M.inputs:
            edges.append((
                '{t3}else:\n'
                '{t4}self._error({input_args})\n').format(
                    t3=3*tab,
                    t4=4*tab,
                    input_args=input_args))
        elif not edges:
            edges.append(
                '{t}raise Exception("Reached dead-end state !")\n'.format(
                    t=3*tab))
        # each state
        c.append((
            '{t2}{ifu} self.state == {node_id}:\n'
            '{edges}').format(
                t2=2*tab,
                ifu=ifu,
                node_id=node_to_int[u],
                edges=''.join(edges)))
    code += ''.join(c) + (
            '{t2}else:\n'
            '{t3}raise Exception("Unrecognized internal state: " + '
            'str(self.state))\n'
            '{t2}return output\n'
            '\n'
            '{t}def _error(self, {input_args}):\n'
            '{t2}raise ValueError("Unrecognized input: " + ('
            '{inputs}).format({args}))\n').format(
                t=tab,
                t2=2*tab,
                t3=3*tab,
                input_args=input_args,
                inputs=''.join(
                    '\n{t}"{v} = {{{v}}}; "'.format(v=v, t=3*tab)
                    for v in M.inputs),
                args=','.join('\n{t}{v}={v}'.format(v=v, t=4*tab)
                              for v in M.inputs))
    return code



# Copyright (c) 2011-2014 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""
Proposition preserving partition module.
"""
from __future__ import absolute_import

import logging
logger = logging.getLogger(__name__)

import warnings
import copy

import numpy as np
from scipy import sparse as sp
import polytope as pc
from polytope.plot import plot_partition

from tulip import transys as trs

# inline imports:
#
# from tulip.graphics import newax

_hl = 40 * '-'

def prop2part(state_space, cont_props_dict):
    """Main function that takes a domain (state_space) and a list of
    propositions (cont_props), and returns a proposition preserving
    partition of the state space.

    See Also
    ========
    L{PropPreservingPartition},
    C{polytope.Polytope}

    @param state_space: problem domain
    @type state_space: C{polytope.Polytope}

    @param cont_props_dict: propositions
    @type cont_props_dict: dict of C{polytope.Polytope}

    @return: state space quotient partition induced by propositions
    @rtype: L{PropPreservingPartition}
    """
    first_poly = [] #Initial Region's polytopes
    first_poly.append(state_space)

    regions = [pc.Region(first_poly)]

    for cur_prop in cont_props_dict:
        cur_prop_poly = cont_props_dict[cur_prop]

        num_reg = len(regions)
        prop_holds_reg = []

        for i in xrange(num_reg): #i region counter
            region_now = regions[i].copy()
            #loop for prop holds
            prop_holds_reg.append(0)

            prop_now = regions[i].props.copy()

            dummy = region_now.intersect(cur_prop_poly)

            # does cur_prop hold in dummy ?
            if pc.is_fulldim(dummy):
                dum_prop = prop_now.copy()
                dum_prop.add(cur_prop)

                # is dummy a Polytope ?
                if len(dummy) == 0:
                    regions[i] = pc.Region([dummy], dum_prop)
                else:
                    # dummy is a Region
                    dummy.props = dum_prop.copy()
                    regions[i] = dummy.copy()
                prop_holds_reg[-1] = 1
            else:
                #does not hold in the whole region
                # (-> no need for the 2nd loop)
                regions.append(region_now)
                continue

            #loop for prop does not hold
            regions.append(pc.Region([], props=prop_now) )
            dummy = region_now.diff(cur_prop_poly)

            if pc.is_fulldim(dummy):
                dum_prop = prop_now.copy()

                # is dummy a Polytope ?
                if len(dummy) == 0:
                    regions[-1] = pc.Region([pc.reduce(dummy)], dum_prop)
                else:
                    # dummy is a Region
                    dummy.props = dum_prop.copy()
                    regions[-1] = dummy.copy()
            else:
                regions.pop()

        count = 0
        for hold_count in xrange(len(prop_holds_reg)):
            if prop_holds_reg[hold_count]==0:
                regions.pop(hold_count-count)
                count+=1

    mypartition = PropPreservingPartition(
        domain = copy.deepcopy(state_space),
        regions = regions,
        prop_regions = copy.deepcopy(cont_props_dict)
    )

    mypartition.adj = pc.find_adjacent_regions(mypartition).copy()

    return mypartition

def part2convex(ppp):
    """This function takes a proposition preserving partition and generates
    another proposition preserving partition such that each part in the new
    partition is a convex polytope

    @type ppp: L{PropPreservingPartition}

    @return: refinement into convex polytopes and
        map from new to old Regions
    @rtype: (L{PropPreservingPartition}, list)
    """
    cvxpart = PropPreservingPartition(
        domain=copy.deepcopy(ppp.domain),
        prop_regions=copy.deepcopy(ppp.prop_regions)
    )
    new2old = []
    for i in xrange(len(ppp.regions)):
        simplified_reg = pc.union(ppp.regions[i],
                                  ppp.regions[i],
                                  check_convex=True)

        for j in xrange(len(simplified_reg)):
            region_now = pc.Region(
                [simplified_reg[j]],
                ppp.regions[i].props
            )
            cvxpart.regions.append(region_now)
            new2old += [i]

    cvxpart.adj = pc.find_adjacent_regions(cvxpart).copy()

    return (cvxpart, new2old)

def pwa_partition(pwa_sys, ppp, abs_tol=1e-5):
    """This function takes:

      - a piecewise affine system C{pwa_sys} and
      - a proposition-preserving partition C{ppp}
          whose domain is a subset of the domain of C{pwa_sys}

    and returns a *refined* proposition preserving partition
    where in each region a unique subsystem of pwa_sys is active.

    Reference
    =========
    Modified from Petter Nilsson's code
    implementing merge algorithm in:

    Nilsson et al.
    `Temporal Logic Control of Switched Affine Systems with an
    Application in Fuel Balancing`, ACC 2012.

    See Also
    ========
    L{discretize}

    @type pwa_sys: L{hybrid.PwaSysDyn}
    @type ppp: L{PropPreservingPartition}

    @return: new partition and associated maps:

        - new partition C{new_ppp}
        - map of C{new_ppp.regions} to C{pwa_sys.list_subsys}
        - map of C{new_ppp.regions} to C{ppp.regions}

    @rtype: C{(L{PropPreservingPartition}, list, list)}
    """
    if pc.is_fulldim(ppp.domain.diff(pwa_sys.domain) ):
        raise Exception('pwa system is not defined everywhere ' +
                        'in state space')

    # for each subsystem's domain, cut it into pieces
    # each piece is the intersection with
    # a unique Region in ppp.regions
    new_list = []
    subsys_list = []
    parents = []
    for i, subsys in enumerate(pwa_sys.list_subsys):
        for j, region in enumerate(ppp.regions):
            isect = region.intersect(subsys.domain)

            if pc.is_fulldim(isect):
                rc, xc = pc.cheby_ball(isect)

                if rc < abs_tol:
                    msg = 'One of the regions in the refined PPP is '
                    msg += 'too small, this may cause numerical problems'
                    warnings.warn(msg)

                # not Region yet, but Polytope ?
                if len(isect) == 0:
                    isect = pc.Region([isect])

                # label with AP
                isect.props = region.props.copy()

                # store new Region
                new_list.append(isect)

                # keep track of original Region in ppp.regions
                parents.append(j)

                # index of subsystem active within isect
                subsys_list.append(i)

    # compute spatial adjacency matrix
    n = len(new_list)
    adj = sp.lil_matrix((n, n), dtype=np.int8)
    for i, ri in enumerate(new_list):
        pi = parents[i]
        for j, rj in enumerate(new_list[0:i]):
            pj = parents[j]

            if (ppp.adj[pi, pj] == 1) or (pi == pj):
                if pc.is_adjacent(ri, rj):
                    adj[i, j] = 1
                    adj[j, i] = 1
        adj[i, i] = 1

    new_ppp = PropPreservingPartition(
        domain = ppp.domain,
        regions = new_list,
        adj = adj,
        prop_regions = ppp.prop_regions
    )
    return (new_ppp, subsys_list, parents)

def add_grid(ppp, grid_size=None, num_grid_pnts=None, abs_tol=1e-10):
    """ This function takes a proposition preserving partition ppp and the size
    of the grid or the number of grids, and returns a refined proposition
    preserving partition with grids.

    Input:

      - `ppp`: a L{PropPreservingPartition} object
      - `grid_size`: the size of the grid,
          type: float or list of float
      - `num_grid_pnts`: the number of grids for each dimension,
          type: integer or list of integer

    Output:

      - A L{PropPreservingPartition} object with grids

    Note: There could be numerical instabilities when the continuous
    propositions in ppp do not align well with the grid resulting in very small
    regions. Performace significantly degrades without glpk.
    """
    if (grid_size!=None)&(num_grid_pnts!=None):
        raise Exception("add_grid: Only one of the grid size or number of \
                        grid points parameters is allowed to be given.")
    if (grid_size==None)&(num_grid_pnts==None):
        raise Exception("add_grid: At least one of the grid size or number of \
                         grid points parameters must be given.")

    dim=len(ppp.domain.A[0])
    domain_bb = ppp.domain.bounding_box
    size_list=list()
    if grid_size!=None:
        if isinstance( grid_size, list ):
            if len(grid_size) == dim:
                size_list=grid_size
            else:
                raise Exception(
                    "add_grid: grid_size isn't given in a correct format."
                )
        elif isinstance( grid_size, float ):
            for i in xrange(dim):
                size_list.append(grid_size)
        else:
            raise Exception("add_grid: "
                "grid_size isn't given in a correct format.")
    else:
        if isinstance( num_grid_pnts, list ):
            if len(num_grid_pnts) == dim:
                for i in xrange(dim):
                    if isinstance( num_grid_pnts[i], int ):
                        grid_size=(
                                float(domain_bb[1][i]) -float(domain_bb[0][i])
                            ) /num_grid_pnts[i]
                        size_list.append(grid_size)
                    else:
                        raise Exception("add_grid: "
                            "num_grid_pnts isn't given in a correct format.")
            else:
                raise Exception("add_grid: "
                    "num_grid_pnts isn't given in a correct format.")
        elif isinstance( num_grid_pnts, int ):
            for i in xrange(dim):
                grid_size=(
                        float(domain_bb[1][i])-float(domain_bb[0][i])
                    ) /num_grid_pnts
                size_list.append(grid_size)
        else:
            raise Exception("add_grid: "
                "num_grid_pnts isn't given in a correct format.")

    j=0
    list_grid=dict()

    while j<dim:
        list_grid[j] = compute_interval(
            float(domain_bb[0][j]),
            float(domain_bb[1][j]),
            size_list[j],
            abs_tol
        )
        if j>0:
            if j==1:
                re_list=list_grid[j-1]
                re_list=product_interval(re_list, list_grid[j])
            else:
                re_list=product_interval(re_list, list_grid[j])
        j+=1

    new_list = []
    parent = []
    for i in xrange(len(re_list)):
        temp_list=list()
        j=0
        while j<dim*2:
            temp_list.append([re_list[i][j],re_list[i][j+1]])
            j=j+2
        for j in xrange(len(ppp.regions)):
            tmp = pc.box2poly(temp_list)
            isect = tmp.intersect(ppp.regions[j], abs_tol)

            #if pc.is_fulldim(isect):
            rc, xc = pc.cheby_ball(isect)
            if rc > abs_tol/2:
                if rc < abs_tol:
                    print("Warning: "
                        "One of the regions in the refined PPP is too small"
                        ", this may cause numerical problems")
                if len(isect) == 0:
                    isect = pc.Region([isect], [])
                isect.props = ppp.regions[j].props.copy()
                new_list.append(isect)
                parent.append(j)

    adj = sp.lil_matrix((len(new_list), len(new_list)), dtype=np.int8)
    for i in xrange(len(new_list)):
        adj[i,i] = 1
        for j in xrange(i+1, len(new_list)):
            if (ppp.adj[parent[i], parent[j]] == 1) or \
                    (parent[i] == parent[j]):
                if pc.is_adjacent(new_list[i], new_list[j]):
                    adj[i,j] = 1
                    adj[j,i] = 1

    return PropPreservingPartition(
        domain = ppp.domain,
        regions = new_list,
        adj = adj,
        prop_regions = ppp.prop_regions
    )

#### Helper functions ####
def compute_interval(low_domain, high_domain, size, abs_tol=1e-7):
    """Helper implementing intervals computation for each dimension.
    """
    list_g=list()
    i=low_domain
    while True:
        if (i+size+abs_tol)>=high_domain:
            list_g.append([i,high_domain])
            break
        else:
            list_g.append([i,i+size])
        i=i+size
    return list_g

def product_interval(list1, list2):
    """Helper implementing combination of all intervals for any two interval lists.
    """
    new_list=list()
    for m in xrange(len(list1)):
        for n in range(len(list2)):
            new_list.append(list1[m]+list2[n])
    return new_list

################################

class PropPreservingPartition(pc.MetricPartition):
    """Partition class with following fields:

      - domain: the domain we want to partition
          type: C{Polytope}

      - regions: Regions of proposition-preserving partition
          type: list of C{Region}

      - adj: a sparse matrix showing which regions are adjacent
          order of C{Region}s same as in list C{regions}

          type: scipy lil sparse

      - prop_regions: map from atomic proposition symbols
          to continuous subsets

          type: dict of C{Polytope} or C{Region}

    See Also
    ========
    L{prop2part}
    """
    def __init__(self,
        domain=None, regions=[],
        adj=None, prop_regions=None, check=True
    ):
        #super(PropPreservingPartition, self).__init__(adj)

        if prop_regions is None:
            self.prop_regions = None
        else:
            try:
                # don't call it
                # use try because it should work
                # vs hasattr, which would look like normal selection
                prop_regions.keys
            except:
                msg = 'prop_regions must be dict.'
                msg += 'Got instead: ' + str(type(prop_regions))
                raise TypeError(msg)

            self.prop_regions = copy.deepcopy(prop_regions)

        n = len(regions)

        if hasattr(adj, 'shape'):
            (m, k) = adj.shape
            if m != k:
                raise ValueError('adj must be square')
            if m != n:
                msg = "adj size doesn't agree with number of regions"
                raise ValueError(msg)

        self.regions = regions[:]

        if check:
            for region in regions:
                if not region <= domain:
                    msg = 'Partition: Region:\n\n' + str(region) + '\n'
                    msg += 'is not subset of given domain:\n\t'
                    msg += str(domain)
                    raise ValueError(msg)

            self.is_symbolic()

        self.domain = domain
        super(PropPreservingPartition, self).__init__(domain)
        self.adj = adj

    def reg2props(self, region_index):
        return self.regions[region_index].props.copy()

    #TODO: iterator over pairs
    #TODO: use nx graph to store partition

    def is_symbolic(self):
        """Check that the set of preserved predicates
        are bijectively mapped to the symbols.

        Symbols = Atomic Propositions
        """
        if self.prop_regions is None:
            msg = 'No continuous propositions defined.'
            logging.warn(msg)
            warnings.warn(msg)
            return

        for region in self.regions:
            if not region.props <= set(self.prop_regions):
                msg = 'Partitions: Region labeled with propositions:\n\t'
                msg += str(region.props) + '\n'
                msg += 'not all of which are in the '
                msg += 'continuous atomic propositions:\n\t'
                msg += str(set(self.prop_regions) )
                raise ValueError(msg)

    def preserves_predicates(self):
        """Return True if each Region <= Predicates for the
        predicates in C{prop_regions.values},
        where C{prop_regions} is a bijection to
        "continuous" propositions of the specification's alphabet.

        Note
        ====
        1. C{prop_regions} in practice need not be injective.
            It doesnt hurt - though creates unnecessary redundancy.

        2. The specification alphabet is fixed an user-defined.
            It should be distinguished from the auxiliary alphabet
            generated automatically during abstraction,
            which defines another partition with
            its own bijection to TS.
        """
        all_props = set(self.prop_regions)

        for region in self.regions:
            # Propositions True in Region
            for prop in region.props:
                preimage = self.prop_regions[prop]

                if not region <= preimage:
                    return False

            # Propositions False in Region
            for prop in all_props.difference(region.props):
                preimage = self.prop_regions[prop]

                if region.intersect(preimage).volume > pc.polytope.ABS_TOL:
                    return False
        return True

    def __str__(self):
        """Get informal string representation."""
        s = '\n' + _hl + '\n'
        s += 'Proposition Preserving Partition:\n'
        s += _hl + 2*'\n'

        s += 'Domain: ' + str(self.domain) + '\n'

        for j, region in enumerate(self.regions):
            s += 'Region: ' + str(j) +'\n'

            if self.prop_regions is not None:
                s += '\t Propositions: '

                active_props = ' '.join(region.props)

                if active_props:
                    s += active_props + '\n'
                else:
                    s += '{}\n'

            s += str(region)

        if hasattr(self.adj, 'todense'):
            s += 'Adjacency matrix:\n'
            s += str(self.adj.todense()) + '\n'
        return s

    def plot(
        self, trans=None, ppp2trans=None, only_adjacent=False,
        ax=None, plot_numbers=True, color_seed=None
    ):
        """For details see C{polytope.plot.plot_partition}.
        """
        return plot_partition(
            self, trans, ppp2trans, only_adjacent,
            ax, plot_numbers, color_seed
        )

    def plot_props(self, ax=None, text_color='yellow'):
        """Plot labeled regions of continuous propositions.
        """
        try:
            from tulip.graphics import newax
        except:
            logger.error('failed to import graphics')
            return

        if ax is None:
            ax, fig = newax()

        l, u = self.domain.bounding_box
        ax.set_xlim(l[0,0], u[0,0])
        ax.set_ylim(l[1,0], u[1,0])

        for (prop, poly) in self.prop_regions.iteritems():
            isect_poly = poly.intersect(self.domain)

            isect_poly.plot(ax, color='none', hatch='/')
            isect_poly.text(prop, ax, color=text_color)
        return ax

class PPP(PropPreservingPartition):
    """Alias to L{PropPreservingPartition}.

    See that for details.
    """
    def __init__(self, **args):
        PropPreservingPartition.__init__(self, **args)

def ppp2ts(part):
    """Derive transition system from proposition preserving partition.

    @param part: labeled polytopic partition from
        which to derive the transition system
    @type part: L{PropPreservingPartition}

    @return: C{(ts, state_map)}
        finite transition system labeled with propositions
        from the given partition, and map of
        polytope indices to transition system states.

    @rtype: (L{transys.FTS}, \C{dict})
    """
    # generate transition system and add transitions
    ofts = trs.FTS()

    adj = part.adj #sp.lil_matrix
    n = adj.shape[0]
    ofts_states = range(n)
    ofts_states = trs.prepend_with(ofts_states, 's')

    ofts.states.add_from(ofts_states)

    ofts.transitions.add_adj(adj, ofts_states)

    # decorate TS with state labels
    atomic_propositions = set(part.prop_regions)
    ofts.atomic_propositions.add_from(atomic_propositions)
    for state, region in zip(ofts_states, part.regions):
        state_prop = region.props.copy()
        ofts.states.add(state, ap=state_prop)

    return (ofts, ofts_states)
